import csv
import os
import shutil
import tempfile
import warnings
import zipfile
from fnmatch import fnmatch
from pathlib import Path

import png
import xmltodict
from build import ProjectBuilder
from jsonschema import validate
from packaging import version

from fme_packager.exception import (
    TransformerPythonCompatError,
    CustomTransformerPythonCompatError,
)
from fme_packager.help import HelpBuilder
from fme_packager.metadata import (
    load_fpkg_metadata,
    load_metadata_json_schema,
    TransformerMetadata,
    FMEPackageMetadata,
)
from fme_packager.operations import (
    build_fpkg_filename,
    parse_formatinfo,
    TREE_COPY_IGNORE_GLOBS,
    load_format_line,
)
from fme_packager.summarizer import summarize_fpkg
from fme_packager.transformer import load_transformer, CustomTransformer


def is_valid_python_compatibility(python_compat_version):
    """
    Checks for a valid python compatibility value.

    :param str python_compat_version: Python compatibility version
    :rtype: bool
    """
    minimum_requirement = "35"
    is_python_3 = python_compat_version.startswith("3")
    meets_minimum_requirements = False
    if is_python_3:
        meets_minimum_requirements = version.parse(python_compat_version) >= version.parse(
            minimum_requirement
        )
    return is_python_3 and meets_minimum_requirements


def check_exists_and_copy(src, dest):
    """
    Copy source contents to a source path.

    :raises ValueError: If the source path does not exist.
    :param src: The source path.
    :param dest: The destination path.
    """
    if not os.path.exists(src):
        raise ValueError("{} is required but does not exist".format(src))
    shutil.copy(src, dest)


def enforce_png(path, min_width=0, min_height=0, square=False):
    """
    Enforces the path is a valid PNG file and the PNG meets a minimum height and width requirement.

    :raises ValueError: If the PNG is invalid or does not meet the minimum requirements.
    :param path: Path to PNG file.
    :param min_width: The minimum width of the image.
    :param min_height: The minimum width of the image.
    :param square: Whether the image should be square.
    """
    width, height, _, _ = png.Reader(filename=path).read()  # Raises if not PNG
    if width < min_width or height < min_height:
        raise ValueError(
            "Min dimensions are {}x{}. {} is {}x{}".format(
                min_width, min_height, path, width, height
            )
        )
    if square and width != height:
        raise ValueError("{} must be square".format(path))


def fq_format_short_name(publisher_uid, package_uid, format_name):
    """
    Formats the publisher_uid, package_uid and format_name

    :param publisher_uid: The Publisher unique identifier.
    :param package_uid: The package unique identifier.
    :param format_name: The format name.
    :return: publisher_uid, package_uid, format_name dot delimited and upper case.
    """
    return "{}.{}.{}".format(publisher_uid, package_uid, format_name).upper()


def check_fmf(package_metadata, format_metadata, fmf_path):
    """
    Checks the fmf file is consistent with the package and format metadata.

    :raises ValueError: If the fmf fails the check.
    :param package_metadata: The package metadata.
    :param format_metadata: The format metadata.
    :param fmf_path: The fmf file path.
    """
    with open(fmf_path) as inf:
        contents = inf.read()

    fqname = fq_format_short_name(
        package_metadata.publisher_uid, package_metadata.uid, format_metadata.name
    )

    if (
        "SOURCE_READER {}".format(fqname) not in contents
        or "FORMAT_NAME {}".format(fqname) not in contents
    ):
        raise ValueError("SOURCE_READER and FORMAT_NAME must be '{}'".format(fqname))


def get_formatinfo(package_metadata, format_metadata, db_path):
    """
    Retrieves format info and checks that it is consistent with the package and format metadata.

    :param package_metadata: The package metadata.
    :param format_metadata: The format metadata.
    :param db_path: The path to the format file.
    """
    fqname = fq_format_short_name(
        package_metadata.publisher_uid, package_metadata.uid, format_metadata.name
    )

    line = load_format_line(db_path)
    if not line:
        raise ValueError("{} empty".format(db_path))

    format_info = parse_formatinfo(line)
    if format_info.FORMAT_NAME == fqname:
        return format_info

    raise ValueError("{} must have FORMAT_NAME of '{}'".format(db_path, fqname))


def get_format_visibility(formatinfo):
    """
    Get the directions for which a format is visible to the user

    :param formatinfo: the format info.
    """
    direction_to_visibility_map = {
        "BOTH": {"YES": "rw", "INPUT": "r", "OUTPUT": "w", "NO": ""},
        "INPUT": {"YES": "r", "INPUT": "r", "NO": ""},  # exclude visibility: OUTPUT
        "OUTPUT": {"YES": "w", "OUTPUT": "w", "NO": ""},  # exclude visibility: INPUT
    }

    try:
        return direction_to_visibility_map[formatinfo.DIRECTION][formatinfo.VISIBLE]
    except KeyError:
        raise ValueError(
            "Format visibility '{visible}' is not compatible with direction '{direction}'".format(
                visible=formatinfo.VISIBLE, direction=formatinfo.DIRECTION
            )
        )


# Enforce unique package_content item names [FMEENGINE-85295]
def enforce_metadata_unique_names(metadata):
    """
    Ensure there are no duplicate item names (case-insensitive)
    within each package_content's items in the given metadata.

    :param metadata: the package metadata.
    :raises ValueError: If a duplicate item name (case-insensitive) is found.
    """

    def validate_unique_names(content_items):
        seen_names = set()
        for item in content_items:
            name = item.name.lower()
            if name in seen_names:
                return name
            seen_names.add(name)
        return None

    content_types = {
        "transformers": metadata.transformers,
        "formats": metadata.formats,
        "web_services": metadata.web_services,
        "web_filesystems": metadata.web_filesystems,
        "python_packages": metadata.python_packages,
    }

    for content_type, content_values in content_types.items():
        duplicate_name = validate_unique_names(content_values)
        if duplicate_name:
            raise ValueError(f"{duplicate_name} is defined in {content_type} more than once.")


def validate_metadata(metadata: FMEPackageMetadata):
    """
    Validate the package metadata against the JSON schema, plus additional checks.
    """
    validate(metadata.dict, load_metadata_json_schema())
    enforce_metadata_unique_names(metadata)
    if (
        metadata.unsupported_from_build
        and metadata.unsupported_from_build <= metadata.minimum_fme_build
    ):
        raise ValueError("unsupported_from_build must be greater than minimum_fme_build")


class FMEPackager:
    def __init__(self, src_dir, verbose=True):
        """
        :param src_dir: FME package directory
        :param verbose: If true, print out build steps
        """
        self.src_dir = Path(src_dir)
        self.build_dir = self.src_dir / "build"
        self.dist_dir = self.src_dir / "dist"
        self.src_python_dir = self.src_dir / "python"

        self.metadata = load_fpkg_metadata(src_dir)
        self.verbose = verbose

        self.fmt_visible_directions = {}

        validate_metadata(self.metadata)

    def apply_help(self, help_src):
        """
        Import an Safe TechPubs doc export into an FME Package directory.

        :param help_src: Help source path.
        """
        tmp_doc_dir = tempfile.mkdtemp(prefix="fme-packager_")
        try:
            # If help source is a ZIP, extract it to a temporary folder.
            if os.path.isfile(help_src):
                with zipfile.ZipFile(help_src) as zipf:
                    self._print("Extracting {} to {}".format(help_src, tmp_doc_dir))
                    zipf.extractall(tmp_doc_dir)
                root_contents = os.listdir(tmp_doc_dir)
                # If help ZIP started with a single root level folder, then unnest.
                if len(root_contents) == 1:
                    nested_dir = os.path.join(tmp_doc_dir, root_contents[0])
                    self._print("Flattening single top-level folder {}".format(nested_dir))
                    for item in os.listdir(nested_dir):
                        shutil.move(os.path.join(nested_dir, item), tmp_doc_dir)
                    os.rmdir(nested_dir)
                help_src = tmp_doc_dir

            # Parse flali file ahead of anything else.
            flali_path = os.path.join(help_src, "package_aliases.flali")
            with open(flali_path, encoding="utf-8") as xmlin:
                aliases_xml = xmltodict.parse(xmlin.read())

            embedded_doc_dirs = [x for x in os.listdir(help_src) if x.startswith("!")]
            if embedded_doc_dirs:
                warnings.warn(
                    "{} embedded doc folders (starting with '!'). Avoid these if possible".format(
                        len(embedded_doc_dirs)
                    )
                )

            # (Re)create destination help folder and copy over the bulk of the doc files.
            dest = os.path.join(self.src_dir, "help")
            if os.path.exists(dest):
                self._print("Deleting {}".format(dest))
                shutil.rmtree(dest)
            shutil.copytree(help_src, dest, ignore=shutil.ignore_patterns(*TREE_COPY_IGNORE_GLOBS))
        finally:
            shutil.rmtree(tmp_doc_dir)

        # Convert flali to CSV and put it in the destination.
        # Skip rows that refer to doc files that aren't present in the doc ZIP.
        copied_rows = 0
        with open(os.path.join(dest, "package_help.csv"), "w", newline="") as csvout:
            writer = csv.writer(csvout)
            rows = aliases_xml["CatapultAliasFile"]["Map"]
            if not isinstance(rows, list):
                rows = [rows]
            for row in rows:
                name = row["@Name"].replace(".", "_")
                link = row["@Link"]
                if link.startswith("/Content"):
                    new_link = link.replace("/Content", "", 1)
                    link = new_link
                expected_doc_path = os.path.join(dest, link.lstrip("/"))
                if os.path.exists(expected_doc_path):
                    self._print("{} exists".format(expected_doc_path))
                    writer.writerow([name, link])
                    copied_rows += 1
        self._print(
            "Wrote {} of {} flali row(s) to package_help.csv".format(copied_rows, len(rows))
        )
        if not copied_rows:
            raise ValueError("flali doesn't reference any included doc")

    def make_fpkg(self):
        """
        Create an .fpkg file from package build files.
        """

        if not self.dist_dir.exists():
            self.dist_dir.mkdir(parents=True)

        fpkg_filename = build_fpkg_filename(
            self.metadata.publisher_uid, self.metadata.uid, self.metadata.version
        )
        fpkg_path = self.dist_dir / fpkg_filename
        self._print(f"Saving fpkg to {fpkg_path}")

        if fpkg_path.exists():
            fpkg_path.unlink()

        new_zip = shutil.make_archive(fpkg_path, "zip", root_dir=self.build_dir)
        Path(new_zip).rename(fpkg_path)  # Strip zip extension
        self._print("Done.")

    def build(self):
        """
        Build and validate package files.
        """
        self._print("Collecting files into {}".format(self.build_dir))

        # Clear out the build dir.
        if os.path.exists(self.build_dir):
            shutil.rmtree(self.build_dir)
        self.build_dir.mkdir(parents=True)

        for required_file in ["package.yml", "README.md", "CHANGES.md"]:
            check_exists_and_copy(self.src_dir / required_file, self.build_dir)

        self._copy_icon()

        self._copy_transformers()
        self._copy_web_services()
        self._copy_web_filesystems()
        self._copy_formats()
        self._copy_localization()
        self._copy_help()

        self._build_wheels()
        self._copy_wheels()
        self._check_wheels()

        # FMEENGINE-87677: Ensure a valid summary can be generated.
        summary = summarize_fpkg(self.build_dir)
        if summary.get("status") == "error":
            raise ValueError(f"Package summary error: {summary.get('message')}")

    def _copy_formats(self):
        # First, copy all files we don't specifically care about.
        # Ignore FMF and etc for formats not mentioned in metadata.
        src = self.src_dir / "formats"
        dst = self.build_dir / "formats"
        if src.is_dir():
            shutil.copytree(src, dst, ignore=shutil.ignore_patterns(*TREE_COPY_IGNORE_GLOBS))

        for fmt in self.metadata.formats:
            self._print(f"Working on format: {fmt.name}")

            if not (dst / f"{fmt.name}.md").is_file():
                raise ValueError(f"{fmt.name} is missing doc")

            fms_path = src / f"{fmt.name}.fms"
            if fms_path.is_file():
                shutil.copy(fms_path, dst)

            fmf_path = src / f"{fmt.name}.fmf"
            if not fmf_path.is_file():
                raise ValueError(f"{fmf_path} is in metadata, but was not found")
            check_fmf(self.metadata, fmt, fmf_path)
            shutil.copy(fmf_path, dst)

            db_path = src / f"{fmt.name}.db"
            if not db_path.is_file():
                raise ValueError(f"{db_path} is in metadata, but was not found")
            formatinfo = get_formatinfo(self.metadata, fmt, db_path)

            self.fmt_visible_directions[fmt.name] = get_format_visibility(formatinfo)

            shutil.copy(db_path, dst)

    def validate_transformer(self, transformer_path, expected_metadata: TransformerMetadata):
        self._print(f"Checking {transformer_path}")
        transformer_file = load_transformer(transformer_path)

        expected_fqname = "{}.{}.{}".format(
            self.metadata.publisher_uid, self.metadata.uid, expected_metadata.name
        )

        transformer_versions = list(transformer_file.versions())
        if not transformer_versions:
            raise ValueError(f"{transformer_path} defines no versions")
        # Validations below apply to all versions of the transformer
        for transformer in transformer_versions:
            if transformer.version < 1:
                raise ValueError(f"Invalid transformer version {transformer.version}")
            if transformer.name != expected_fqname:
                raise ValueError(f"Name must be '{expected_fqname}', not '{transformer.name}'")
            if isinstance(transformer, CustomTransformer):
                if transformer.header.insert_mode != "Linked Always":
                    raise ValueError(
                        f'Custom transformer Insert Mode must be "Linked Always", not {transformer.header.insert_mode}'
                    )
                if transformer.header.build_num < self.metadata.minimum_fme_build:
                    raise ValueError(
                        "Custom transformer created with FME build older than fme_minimum_build in package.yml"
                    )

        # Validations below apply only to the latest version of the transformer
        latest = transformer_versions[0]  # First definition in file
        if latest.version != expected_metadata.version:
            raise ValueError(f"Missing version {expected_metadata.version}")
        if not latest.python_compatibility and self.metadata.minimum_fme_build < 22000:
            # we only care about PYTHON_COMPATIBILITY before Python 2 support was removed (b22000)
            raise ValueError("Python Compatibility setting required when minimum_fme_build < 22000")
        elif latest.python_compatibility and not is_valid_python_compatibility(
            latest.python_compatibility
        ):
            if not isinstance(latest, CustomTransformer):
                raise TransformerPythonCompatError(expected_metadata.name)
            elif isinstance(latest, CustomTransformer) and latest.python_compatibility != "2or3":
                raise CustomTransformerPythonCompatError(expected_metadata.name)

    def _copy_transformers(self):
        # First, copy all files we don't specifically care about.
        # Ignore FMX and FMS for transformers not mentioned in metadata.
        src = self.src_dir / "transformers"
        dst = self.build_dir / "transformers"
        if src.is_dir():
            shutil.copytree(src, dst, ignore=shutil.ignore_patterns(*TREE_COPY_IGNORE_GLOBS))

        for transformer in self.metadata.transformers:
            self._print(f"Working on transformer: {transformer.name}")

            if transformer.version < 1:
                raise ValueError(f"{transformer.name} version must be >= 1")

            if not (dst / f"{transformer.name}.md").is_file():
                raise ValueError(f"{transformer.name} is missing doc")

            fms_path = src / f"{transformer.name}.fms"
            if fms_path.is_file():
                shutil.copy(fms_path, dst)

            fmx_path = src / f"{transformer.name}.fmx"
            fmxj_path = src / f"{transformer.name}.fmxj"
            if fmx_path.is_file() and fmxj_path.is_file():
                raise ValueError(f"{transformer.name} cannot have both FMX and FMXJ")
            if not fmx_path.is_file() and not fmxj_path.is_file():
                raise ValueError(f"{fmx_path} is in metadata, but was not found")
            transformer_path = fmx_path if fmx_path.is_file() else fmxj_path
            if transformer_path == fmxj_path and self.metadata.minimum_fme_build < 23146:
                raise ValueError("FMXJ requires minimum_fme_build >= 23146")

            # Copy file if it passed validation
            self.validate_transformer(transformer_path, transformer)
            shutil.copy(transformer_path, dst)

    def _copy_web_services(self):
        dest = self.build_dir / "web_services"
        for web_service in self.metadata.web_services:
            # fpkg spec says web service metadata entries must be full filename. Don't assume xml.
            definition_path = self.src_dir / "web_services" / web_service.name
            if not definition_path.exists():
                raise ValueError(
                    f"Web Service '{web_service.name}' is in metadata, but was not found"
                )
            if not dest.exists():
                dest.mkdir(parents=True)

            # TODO: Validate contents of XML.
            self._print(f"Copying Web Service: {definition_path.name}")
            shutil.copy(definition_path, dest)

    def _copy_web_filesystems(self):
        src = self.src_dir / "web_filesystems"
        dest = self.build_dir / "web_filesystems"
        for web_filesystem in self.metadata.web_filesystems:
            definition_path = src / f"{web_filesystem.name}.fme"
            if not definition_path.is_file():
                raise ValueError(
                    f"Web Filesystem '{web_filesystem.name}' is in metadata, but was not found"
                )
            if not dest.exists():
                dest.mkdir(parents=True)

            # TODO: Validate contents of .fme file.
            self._print(f"Copying Web Filesystem: {definition_path.name}")
            shutil.copy(definition_path, dest)

            icon_path = src / f"{web_filesystem.name}.png"
            if icon_path.is_file():
                enforce_png(icon_path)
                # Specification doesn't say anything about dimensions.
                shutil.copy(icon_path, dest)

    def _copy_icon(self):
        path = self.src_dir / "icon.png"
        if not path.is_file():
            self._print("FME package has no icon")
            return
        enforce_png(path, min_width=200, min_height=200, square=True)
        self._print("Icon is OK")
        shutil.copy(path, self.build_dir)

    def _build_wheels(self):
        original_cwd = os.getcwd()

        if not self.src_python_dir.exists():
            return

        files = ("pyproject.toml", "setup.cfg", "setup.py")
        for path in self.src_python_dir.iterdir():
            # If any of the expected files are present, build the wheel.
            if any((path / file).is_file() for file in files):
                self._print(f"\nBuilding Python wheel: {path}")
                os.chdir(path)
                if os.path.exists("build"):
                    shutil.rmtree("build")
                # Simple solution to avoid needing to figure out which output is the one we want.
                # The only wheel in the directory will be the one to include in the fpkg.
                if os.path.exists("dist"):
                    shutil.rmtree("dist")
                try:
                    result = ProjectBuilder(".").build("wheel", "dist")
                    print(f"Built {result}")
                finally:
                    os.chdir(original_cwd)

    def _copy_wheels(self):
        if not self.src_python_dir.is_dir():
            return

        wheels_dest = self.build_dir / "python"
        wheels_dest.mkdir(parents=True)

        for path in self.src_python_dir.iterdir():
            if os.path.isfile(path) and path.suffix == ".whl":
                shutil.copy(path, wheels_dest)

            built_wheels_dir = path / "dist"
            if built_wheels_dir.is_dir():
                built_wheels_for_lib = [
                    name for name in built_wheels_dir.iterdir() if name.suffix == ".whl"
                ]
                assert len(built_wheels_for_lib) == 1
                shutil.copy(built_wheels_for_lib[0], wheels_dest)

    def _check_wheels(self):
        wheels_path = self.build_dir / "python"

        wheel_names = os.listdir(wheels_path) if wheels_path.is_dir() else []
        for expected_py_package in self.metadata.python_packages:
            lib_name = expected_py_package.name
            if not any(
                wheel_name.startswith(lib_name) or wheel_name.startswith(lib_name.replace("-", "_"))
                for wheel_name in wheel_names
            ):
                raise ValueError(
                    f"Python library '{lib_name}' is in metadata but not in {', '.join(wheel_names)}"
                )

        if not wheels_path.is_dir():
            return
        for wheel_path in wheels_path.iterdir():
            wheel_name = wheel_path.name
            if "py3" not in wheel_name:
                warnings.warn("{} is not a Python 3 wheel".format(wheel_name))
            if not wheel_name.endswith("-none-any.whl"):
                warnings.warn("{} is not a pure-Python wheel".format(wheel_name))

    def _copy_localization(self):
        # Copy any optional localization files. The only sanity check is a filename whitelist:
        # guiprompts_??.txt, transformer-localized.??, FormatOrTransformerName_??.md
        # ?? is a 2-letter language code.
        # Contents of files aren't validated. It's up to Workbench to ignore anything malformed.
        dest = self.build_dir / "localization"
        src = self.src_dir / "localization"
        if not src.is_dir():
            return

        keywords = [f.name for f in self.metadata.formats] + [
            t.name for t in self.metadata.transformers
        ]
        localized_doc_globs = [keyword + "_??.md" for keyword in keywords]

        for name in os.listdir(src):
            path = src / name
            if (
                fnmatch(name, "guiprompts_??.txt")
                or fnmatch(name, "transformer-localized.??")
                or any(fnmatch(name, glob) for glob in localized_doc_globs)
            ):
                if not os.path.exists(dest):
                    os.makedirs(dest)
                self._print(f"Copying localization: {name}")
                shutil.copy(path, dest)

    def _copy_help(self):
        src = self.src_dir / "help"
        dest = self.build_dir / "help"
        if not src.is_dir():
            return

        self._print("Copying help")
        builder = HelpBuilder(self.metadata, src, dest, self.fmt_visible_directions)
        builder.build()

    def _print(self, msg):
        if self.verbose:
            print(msg)



import os
import pathlib
from zipfile import ZipFile
from copy import deepcopy

import pytest
from click.testing import CliRunner
from jsonschema import ValidationError
from ruamel.yaml import YAML

from fme_packager.cli import pack
from fme_packager.exception import (
    TransformerPythonCompatError,
    CustomTransformerPythonCompatError,
)
from fme_packager.metadata import TransformerMetadata, FormatMetadata, FMEPackageMetadata
from fme_packager.operations import parse_formatinfo
from fme_packager.packager import (
    FMEPackager,
    is_valid_python_compatibility,
    get_formatinfo,
    get_format_visibility,
    validate_metadata,
)
from fme_packager.metadata import load_fpkg_metadata

CWD = pathlib.Path(__file__).parent.resolve()


@pytest.mark.parametrize(
    "version, expected_is_valid",
    [
        ("26", False),
        ("27", False),
        ("ArcGISDesktop", False),
        ("34", False),
        ("35", True),
        ("36", True),
        ("37", True),
        ("3.8.9", False),  # 3.x.x is invalid version syntax for PYTHON_COMPATIBILITY
        ("2or3", False),  # only valid for custom transformers
    ],
)
def test_is_valid_python_compatibility(version, expected_is_valid):
    assert is_valid_python_compatibility(version) == expected_is_valid


def test_get_formatinfo(tmp_path):
    package_metadata = FMEPackageMetadata(
        {
            "uid": "package",
            "publisher_uid": "example",
            "package_content": {"formats": [{"name": "myformat"}]},
        }
    )
    format_metadata = FormatMetadata({"name": "myformat"})

    with pytest.raises(ValueError):
        filepath = tmp_path / "emptyformat.db"
        with open(filepath, "w") as f:
            f.write(";")
        get_formatinfo(package_metadata, format_metadata, filepath)

    with pytest.raises(ValueError):
        filepath = tmp_path / "formatnotfound.db"
        with open(filepath, "w") as f:
            f.write(
                "SAFE.CKAN.CKANDATASTORE|CKAN DataStore|NONE|BOTH|NONE|NO||NON_SPATIAL|NO|YES|YES|YES|3|3|CKAN|NO|CKAN"
            )
        get_formatinfo(package_metadata, format_metadata, filepath)

    filepath = tmp_path / "myformat.db"
    with open(filepath, "w") as f:
        f.writelines(
            "EXAMPLE.PACKAGE.MYFORMAT|My Format|NONE|BOTH|NONE|NO||NON_SPATIAL|NO|YES|YES|YES|3|3|MYFORMAT|NO|MYFORMAT|Coordinates\n"
            + "EXAMPLE.PACKAGE.NOTMYFORMAT|Not My Format|NONE|BOTH|NONE|NO||NON_SPATIAL|NO|YES|YES|YES|3|3|MYFORMAT|NO|MYFORMAT|Coordinates,3D"
        )

    formatinfo = get_formatinfo(package_metadata, format_metadata, filepath)
    assert formatinfo.FORMAT_NAME == "EXAMPLE.PACKAGE.MYFORMAT"


def test_get_format_visibility():
    base_formatinfo = parse_formatinfo(
        "EXAMPLE.PACKAGE.MYFORMAT|My Format|NONE|BOTH|NONE|NO||NON_SPATIAL|NO|YES|YES|YES|3|3|MYFORMAT|NO|MYFORMAT|Coordinates"
    )

    assert get_format_visibility(base_formatinfo._replace(DIRECTION="BOTH", VISIBLE="NO")) == ""
    assert get_format_visibility(base_formatinfo._replace(DIRECTION="BOTH", VISIBLE="YES")) == "rw"

    assert get_format_visibility(base_formatinfo._replace(DIRECTION="BOTH", VISIBLE="INPUT")) == "r"
    for visibility in ["YES", "INPUT"]:
        assert (
            get_format_visibility(base_formatinfo._replace(DIRECTION="INPUT", VISIBLE=visibility))
            == "r"
        )

    assert (
        get_format_visibility(base_formatinfo._replace(DIRECTION="BOTH", VISIBLE="OUTPUT")) == "w"
    )
    for visibility in ["YES", "OUTPUT"]:
        assert (
            get_format_visibility(base_formatinfo._replace(DIRECTION="OUTPUT", VISIBLE=visibility))
            == "w"
        )

    with pytest.raises(ValueError):
        assert get_format_visibility(base_formatinfo._replace(DIRECTION="INPUT", VISIBLE="OUTPUT"))


@pytest.mark.parametrize(
    "transformer_path,metadata,expected_exc",
    [
        ("valid_package/transformers/MyGreeter.fmx", None, None),
        (
            "valid_package/transformers/MyGreeter.fmx",
            {"name": "TheirGreeter", "version": 1},
            "Name must be",
        ),
        ("incompatible_package/transformers/MyGreeter.fmx", None, TransformerPythonCompatError),
        (
            "incompatible_custom_package/transformers/epochToTimestamp29.fmx",
            None,
            CustomTransformerPythonCompatError,
        ),
        ("custom_package/transformers/customFooBar.fmx", None, None),
        ("custom_package/transformers/epochToTimestamp29.fmx", None, None),
        (
            "custom_package/transformers/customEncrypted2Ver.fmx",
            {"name": "test", "version": 2},
            'Custom transformer Insert Mode must be "Linked Always"',
        ),
        ("fmxj_package/transformers/DemoGreeter.fmxj", None, None),
    ],
)
def test_validate_transformer(transformer_path, metadata, expected_exc):
    """
    Load and validate a transformer.
    If expected_exc is given and is an exception, it's expected to be raised.
    If it's a string, it's expected to be a substring in the raised exception message.
    The package.yml in the transformer's parent folder is automatically loaded.
    The metadata argument overrides the transformer's entry in package.yml.
    If the transformer has no entry in package.yml, the metadata must be provided.
    """
    transformer_abs_path = CWD / "fixtures" / transformer_path
    packager = FMEPackager(transformer_abs_path.parent.parent)
    if metadata:
        metadata = TransformerMetadata(metadata)
    else:
        try:
            metadata = {i.name: i for i in packager.metadata.transformers}[
                transformer_abs_path.stem
            ]
        except KeyError as e:
            raise KeyError(f"{transformer_abs_path.stem} not in package.yml") from e

    if not expected_exc:
        # Expect success
        packager.validate_transformer(transformer_abs_path, metadata)
    elif isinstance(expected_exc, str):
        # Expect a keywords in exception message
        with pytest.raises(Exception) as exc:
            packager.validate_transformer(transformer_abs_path, metadata)
        assert expected_exc in str(exc.value)
    else:
        # Expect a particular exception class
        with pytest.raises(expected_exc):
            packager.validate_transformer(transformer_abs_path, metadata)


@pytest.mark.parametrize("package_name", ["valid_package", "fmxj_package"])
def test_pack(package_name):
    runner = CliRunner()
    result = runner.invoke(pack, [str(CWD / "fixtures" / package_name)])
    assert result.exit_code == 0

    # Sanity check: expect FPKG to exist and contain package.yml at root level
    dist_dir = CWD / "fixtures" / package_name / "dist"
    fpkg_name = os.listdir(dist_dir)[0]
    with ZipFile(dist_dir / fpkg_name) as z:
        assert "package.yml" in z.namelist()


PACKAGE_YML_TEMPLATE = {
    "uid": "test-uid",
    "publisher_uid": "test-publisher",
    "name": "test-name",
    "description": "test_description",
    "author": {
        "name": "name",
        "email": "a@a.com",
    },
    "version": "1.0.0",
    "minimum_fme_build": 25000,
    "fpkg_version": 1,
    "package_content": {
        "transformers": [
            {"name": "MyTransformer", "version": 1},
        ],
        "formats": [
            {"name": "MyFormat"},
        ],
        "web_services": [
            {"name": "MyWebService.xml"},
        ],
    },
}


def _dupe_content(yml):
    # Add a copy of the first element to each package_content sublist
    for key, value in yml["package_content"].items():
        if not value:
            continue
        value.append(value[0])


PACKAGE_YML_VALIDATION_CASES = [
    pytest.param(None, None, id="valid"),
    pytest.param(_dupe_content, "has non-unique elements", id="dupe_content"),
    pytest.param(
        lambda x: x.update(
            {
                "package_content": {
                    "transformers": [
                        {"name": "XFormer", "version": 1},
                        {"name": "XFormer", "version": 0},
                    ]
                }
            }
        ),
        "xformer is defined in transformers more than once",
        id="dupe_content_name_only",  # FMEENGINE-85295
    ),
    pytest.param(
        lambda x: x["package_content"].update({}),
        None,
        id="empty_content_valid",
    ),
    # https://github.com/python-jsonschema/jsonschema/commit/a1d4cb3b94a83f209ebde1506a18b297a42b6219
    pytest.param(
        lambda x: x["package_content"].update({"transformers": []}),
        "[] is too short",
        id="empty_content_subkey_lt_py_311",
        marks=pytest.mark.skipif(
            os.sys.version_info >= (3, 11),
            reason="jsonschema pinned on PY<3.11, which has different msg",
        ),
    ),
    pytest.param(
        lambda x: x["package_content"].update({"transformers": []}),
        "[] should be non-empty",
        id="empty_content_subkey",
        marks=pytest.mark.skipif(
            os.sys.version_info < (3, 11),
            reason="jsonschema pinned on PY<3.11, which has different msg",
        ),
    ),
    pytest.param(
        lambda x: x.update({"version": 1}),
        "Failed validating 'type'",
        id="version_not_str",
    ),
    pytest.param(
        lambda x: x.update({"version": "a.b.c"}),
        "Failed validating 'pattern'",
        id="version_not_semantic",
    ),
    pytest.param(
        lambda x: x.update({"foo": "bar"}),
        "Additional properties are not allowed",
        id="extra_root_key",
    ),
    pytest.param(
        lambda x: x.update({"unsupported_from_build": 25000}),
        "25000 is less than the minimum",
        id="unsupported_from_build_too_low",
    ),
    pytest.param(
        lambda x: x.update({"minimum_fme_build": 26000, "unsupported_from_build": 26000}),
        "unsupported_from_build must be greater than minimum_fme_build",
        id="unsupported_from_build_lte_minimum_fme_build",
    ),
    pytest.param(
        lambda x: x.update({"minimum_fme_build": 26000, "unsupported_from_build": 26001}),
        None,
        id="unsupported_from_build_valid",
    ),
]
for required_key in (
    "uid",
    "publisher_uid",
    "name",
    "description",
    "author",
    "version",
    "minimum_fme_build",
    "fpkg_version",
    "package_content",
):
    PACKAGE_YML_VALIDATION_CASES.append(
        pytest.param(
            lambda x, key=required_key: x.pop(key),
            f"'{required_key}' is a required property",
            id=f"missing_{required_key}",
        )
    )
for key in ("uid", "publisher_uid"):
    PACKAGE_YML_VALIDATION_CASES.append(
        pytest.param(
            lambda x, key=key: x.__setitem__(key, "a_b"),
            "Failed validating 'pattern'",
            id=f"bad_{key}",
        )
    )


@pytest.mark.parametrize("mutation_func, expect_msg", PACKAGE_YML_VALIDATION_CASES)
def test_package_yml_validation(mutation_func, expect_msg, tmp_path):
    """
    Starting with a valid package.yml template, apply mutations and validate.
    If expect_msg is given, it's expected to be a substring in the raised exception message.
    If not given, validation is expected to succeed.
    """
    package_yml_dict = deepcopy(PACKAGE_YML_TEMPLATE)
    if mutation_func:
        mutation_func(package_yml_dict)

    with (tmp_path / "package.yml").open("w") as f:
        YAML().dump(package_yml_dict, f)
    metadata = load_fpkg_metadata(tmp_path)

    if expect_msg:
        with pytest.raises((ValidationError, ValueError)) as exc:
            validate_metadata(metadata)
        assert expect_msg in str(exc.value)
    else:
        # Expect success
        validate_metadata(metadata)



from importlib.resources import open_text
import json
import os
from pathlib import Path
from typing import Optional, Union
from functools import cache

from ruamel.yaml import YAML


def load_fpkg_metadata(fpkg_path: Union[str, os.PathLike]):
    """
    Load package.yml as a FMEPackageMetadata object.

    :param fpkg_path: Path to the directory where package.yml exists.
    :rtype: FMEPackageMetadata
    """
    with open(Path(fpkg_path) / "package.yml") as f:
        return FMEPackageMetadata(YAML(typ="safe").load(f))


@cache
def load_metadata_json_schema() -> dict:
    """
    Load in package metadata specification and requirements.
    """
    with open_text("fme_packager", "spec.json") as f:
        return json.load(f)


class TransformerMetadata:
    def __init__(self, metadata_dict):
        self.dict = metadata_dict

    @property
    def name(self):
        return self.dict["name"]

    @property
    def version(self):
        return self.dict["version"]


class PythonPackageMetadata:
    def __init__(self, metadata_dict):
        self.dict = metadata_dict

    @property
    def name(self):
        return self.dict["name"]


class FormatMetadata:
    def __init__(self, metadata_dict):
        self.dict = metadata_dict

    @property
    def name(self):
        return self.dict["name"]


class WebServiceMetadata:
    def __init__(self, metadata_dict):
        self.dict = metadata_dict

    @property
    def name(self):
        return self.dict["name"]


class WebFilesystemMetadata:
    def __init__(self, metadata_dict):
        self.dict = metadata_dict

    @property
    def name(self):
        return self.dict["name"]


class Author:
    def __init__(self, metadata_dict):
        self.dict = metadata_dict

    @property
    def name(self):
        return self.dict["name"]

    @property
    def email(self):
        return self.dict["name"]


class FMEPackageMetadata:
    def __init__(self, metadata_dict):
        self.dict = metadata_dict

    @property
    def fpkg_version(self):
        return self.dict["fpkg_version"]

    @property
    def uid(self):
        return self.dict["uid"]

    @property
    def publisher_uid(self):
        return self.dict["publisher_uid"]

    @property
    def name(self):
        return self.dict["name"]

    @property
    def description(self):
        return self.dict["description"]

    @property
    def version(self):
        return self.dict["version"]

    @property
    def minimum_fme_build(self):
        return self.dict["minimum_fme_build"]

    @property
    def unsupported_from_build(self) -> Optional[int]:
        return self.dict.get("unsupported_from_build", None)

    @property
    def author(self):
        return Author(self.dict["author"])

    @property
    def package_content(self):
        return self.dict.get("package_content", {})

    @property
    def transformers(self):
        return [TransformerMetadata(item) for item in self.package_content.get("transformers", [])]

    @property
    def python_packages(self):
        return [
            PythonPackageMetadata(item) for item in self.package_content.get("python_packages", [])
        ]

    @property
    def formats(self):
        return [FormatMetadata(item) for item in self.package_content.get("formats", [])]

    @property
    def web_services(self):
        return [WebServiceMetadata(item) for item in self.package_content.get("web_services", [])]

    @property
    def web_filesystems(self):
        return [
            WebFilesystemMetadata(item) for item in self.package_content.get("web_filesystems", [])
        ]



import pathlib

import pytest

from fme_packager.help import HelpBuilder, get_expected_help_index
from fme_packager.metadata import FMEPackageMetadata

CWD = pathlib.Path(__file__).parent.resolve()
HELP_FIXTURES_DIR = CWD / "fixtures" / "help"


@pytest.fixture
def mock_metadata():
    return FMEPackageMetadata(
        {
            "uid": "package-hyphen",
            "publisher_uid": "example",
            "minimum_fme_build": 23224,
            "package_content": {"transformers": [{"name": "Transformer", "version": 1}]},
        }
    )


def test_get_expected_help_contexts_transformer(mock_metadata):
    assert get_expected_help_index(mock_metadata)[0] == {
        "fmx_example_package-hyphen_Transformer": "/Transformer.htm"
    }


def test_get_expected_help_contexts_format():
    metadata = FMEPackageMetadata(
        {
            "uid": "package-hyphen",
            "publisher_uid": "example",
            "package_content": {"formats": [{"name": "demoformat"}]},
        }
    )
    assert get_expected_help_index(metadata)[0] == {
        "ft_example_package_hyphen_demoformat_param_r": "/demoformat_ft_param_r.htm",
        "ft_example_package_hyphen_demoformat_param_w": "/demoformat_ft_param_w.htm",
        "ft_example_package_hyphen_demoformat_user_attr": "/demoformat_ft_user_attr.htm",
        "param_example_package_hyphen_demoformat_r": "/demoformat_param_r.htm",
        "param_example_package_hyphen_demoformat_w": "/demoformat_param_w.htm",
        "rw_example_package_hyphen_demoformat_feature_rep": "/demoformat_feature_rep.htm",
        "rw_example_package_hyphen_demoformat_index": "/demoformat.htm",
    }
    assert get_expected_help_index(metadata)[1] == {
        "rw_example_package_hyphen_demoformat_quickfacts": "/demoformat_quickfacts.htm",
    }
    assert sorted(get_expected_help_index(metadata)[0]) == [
        "ft_example_package_hyphen_demoformat_param_r",
        "ft_example_package_hyphen_demoformat_param_w",
        "ft_example_package_hyphen_demoformat_user_attr",
        "param_example_package_hyphen_demoformat_r",
        "param_example_package_hyphen_demoformat_w",
        "rw_example_package_hyphen_demoformat_feature_rep",
        "rw_example_package_hyphen_demoformat_index",
    ]
    assert sorted(get_expected_help_index(metadata)[1]) == [
        "rw_example_package_hyphen_demoformat_quickfacts",
    ]


def test_get_expected_help_contexts_format_one_dir():
    metadata = FMEPackageMetadata(
        {
            "uid": "package",
            "publisher_uid": "example",
            "package_content": {"formats": [{"name": "demoformat"}]},
        }
    )
    read_only = {"demoformat": "r"}
    write_only = {"demoformat": "w"}
    assert sorted(get_expected_help_index(metadata, read_only)[0]) == [
        "ft_example_package_demoformat_param_r",
        "ft_example_package_demoformat_user_attr",
        "param_example_package_demoformat_r",
        "rw_example_package_demoformat_feature_rep",
        "rw_example_package_demoformat_index",
    ]
    assert sorted(get_expected_help_index(metadata, write_only)[0]) == [
        "ft_example_package_demoformat_param_w",
        "ft_example_package_demoformat_user_attr",
        "param_example_package_demoformat_w",
        "rw_example_package_demoformat_feature_rep",
        "rw_example_package_demoformat_index",
    ]
    assert sorted(get_expected_help_index(metadata, read_only)[1]) == [
        "rw_example_package_demoformat_quickfacts",
    ]


def test_htm(mock_metadata, tmp_path):
    maker = HelpBuilder(mock_metadata, HELP_FIXTURES_DIR / "htm", tmp_path, {})
    maker.build()
    assert (tmp_path / "Transformers" / "Transformer-pkg.htm").is_file()
    assert (tmp_path / "package_help.csv").is_file()


def test_md(mock_metadata, tmp_path):
    maker = HelpBuilder(mock_metadata, HELP_FIXTURES_DIR / "md", tmp_path, {})
    maker.build()
    assert not (tmp_path / "Transformer.md").is_file()
    html_file = tmp_path / "Transformer.htm"
    with html_file.open("r") as f:
        html = f.read()
        assert html.startswith("<!DOCTYPE html>")
        assert '<link rel="stylesheet" href="../../css/style.css" />' in html
        assert '<h1 class="fmx">' in html
        assert '<p><span class="TransformerSummary">' in html
    index_file = tmp_path / "package_help.csv"
    assert index_file.is_file()
    with index_file.open("r") as f:
        assert f.read() == "fmx_example_package-hyphen_Transformer,/Transformer.htm\n"



import xml
from pathlib import Path

import xmltodict


def _web_service_path(web_service_file_name: str) -> Path:
    return Path("web_services") / f"{web_service_file_name}"


def _parse_web_service(path: Path) -> dict:
    try:
        with open(path, encoding="utf-8") as web_service:
            package_xml = xmltodict.parse(web_service.read())

        webservice_data = package_xml.get("ImportExportData", {}).get("webservices", {})

        # 2024.1 and earlier
        webservice_xml = webservice_data.get("webservicexml", "")

        if not webservice_data.get("webservice") and webservice_xml:
            webservice_data = xmltodict.parse(webservice_xml)

        return webservice_data.get("webservice", {}).get("authentication", {})

    except (xml.parsers.expat.ExpatError, KeyError):
        return {}



import os
import shutil
from collections import namedtuple
from typing import Union


def split_fpkg_filename(filename):
    """
    Split a filename of the form `A.B-C.fpkg`.

    :param filename: Filename to split.
    :raises ValueError: If filename doesn't end in "fpkg", or any resulting value is empty.
    :return: publisher_uid, package_uid, version.
        These values aren't validated, but are guaranteed to not be empty.
    """
    name, _, extension = filename.partition(".")
    if extension != "fpkg":
        raise ValueError("FME Package extension must be 'fpkg', not '{}'".format(extension))
    publisher_uid, _, right = name.partition(".")
    package_uid, _, version = right.rpartition("-")
    if not publisher_uid or not package_uid or not version:
        raise ValueError("Unrecognized fpkg name '{}'".format(name))
    return publisher_uid, package_uid, version


def build_fpkg_filename(publisher_uid, package_uid, version):
    """
    Build an fpkg filename. Inputs aren't validated, but must not be empty.

    :return: Assembled fpkg filename.
    """
    if not publisher_uid or not package_uid or not version:
        raise ValueError("All parts of the fpkg filename are required")
    return "{}.{}-{}.fpkg".format(publisher_uid, package_uid, version)


FORMATINFO_HDR = "FORMAT_NAME|FORMAT_LONG_NAME|DATASET_TYPE|DIRECTION|AUTOMATED_TRANSLATION_FLAG|COORDSYS_AWARE|FILTER|FORMAT_TYPE|USE_NATIVE_SPATIAL_INDEX|SOURCE_SETTINGS|DESTINATION_SETTINGS|VISIBLE|MIN_VERSION|MAX_VERSION|FORMAT_FAMILY|HAS_SIDECARS|MARKETING_FAMILY|FORMAT_CATEGORIES"
FormatInfo = namedtuple("FormatInfo", FORMATINFO_HDR.replace("|", " "), defaults=[""])
OPTIONAL_FORMATINFO_COLUMNS = ["FORMAT_CATEGORIES"]


def parse_formatinfo(line) -> FormatInfo:
    """
    Parse format info line with the FORMATINFO_HDR.

    :param str line: raw format info.
    :return: Parsed format into a FormatInfo named tuple.
    """
    num_columns = len(FORMATINFO_HDR.split("|"))
    num_required_columns = num_columns - len(OPTIONAL_FORMATINFO_COLUMNS)
    parts = line.strip().split("|")
    if len(parts) not in [num_columns, num_required_columns]:
        raise ValueError(
            "FormatInfo has {} elements, but expects at least {}".format(
                len(parts), num_required_columns
            )
        )
    return FormatInfo(*parts)


# Matches files that should not automatically be copied into an fpkg.
# Some of these are OS-specific metadata.
# FME file extensions are here because their copy over is gated by validation.
TREE_COPY_IGNORE_GLOBS = [
    ".*",
    "*.mclog",
    "*.flali",
    "*.fmf",
    "*.fmx",
    "*.db",
    "*.fms",
    ".DS_Store",
    "Thumbs.db",
    "desktop.ini",
]


def extract_fpkg(
    fpkg_file_path: Union[str, os.PathLike], dest_dir: Union[str, os.PathLike]
) -> None:
    """
    Extract .fpkg file to a directory.
    """
    shutil.unpack_archive(fpkg_file_path, dest_dir, "zip")


def load_format_line(db_path: str) -> str:
    """Get the DB info line from the format file."""
    with open(db_path) as inf:
        for line in inf:
            if line.startswith(";"):
                continue  # comment line
            return line.rstrip()

    return ""



"""
Logic for preparing help documentation for the FPKG packing process.

- package_help.csv validation
- File copy logic for the help folder
- Markdown to HTML conversion and package_help.csv generation
"""

import csv
import re
import shutil
import warnings
from pathlib import Path

from markdown import Markdown

from fme_packager.metadata import FMEPackageMetadata
from fme_packager.operations import TREE_COPY_IGNORE_GLOBS


HTML_TPL = """<!DOCTYPE html>
<html>
<head>
<title>{title}</title>
<link rel="stylesheet" href="{css_path}" />
</head>
<body>
{body}
</body>
</html>
"""


def add_transformer_classes_to_doc(html):
    """
    Given a string that is HTML doc, modify the first h1 and p tags
    to mark the relevant contents to show for the FME Workbench Quick Add pane.
    See FMEENGINE-75813.

    Assumes that:
    - these tags don't already have a class attribute
    - the HTML was generated by python-markdown without extensions that modify tags
    """
    # Must be exact match: Workbench finds these tags through text search, not DOM.
    html, _ = re.subn(r"<h1>", '<h1 class="fmx">', html, count=1)
    html, _ = re.subn(r"<p>", '<p><span class="TransformerSummary">', html, count=1)
    html, _ = re.subn(r"</p>", "</span></p>", html, count=1)
    return html


class DocCopier:
    """
    Callable with the same signature as shutil.copy2(), for use with shutil.copytree().
    Copies files while converting MD files to HTML at the destination.
    Identical to shutil.copy2() when conversion is disabled.
    """

    def __init__(self, root_path):
        """
        :param root_path: All files copied must be descendants of this root.
            Used to determine hardcoded relative paths in generated HTML.
        """
        self.root_path = root_path
        self.convert_md = True
        self.converted_files = {}
        self._md_converter = Markdown(
            extensions=[
                "tables",
                "fenced_code",
            ],
        )
        # DEVOPS-3078: Path to CSS relative to the doc root folder
        self._css_path_rel_doc = "../../css/style.css"

        # When generating HTML, transformer doc gets special treatment.
        self.transformer_names = set()

    def md_to_html(self, src_file: Path):
        # Not using python-markdown extensions API because this is a one-off task.
        body = self._md_converter.reset().convert(src_file.read_text("utf8"))
        if src_file.stem in self.transformer_names:
            body = add_transformer_classes_to_doc(body)
        # Count number of subfolders down from the root,
        # so the CSS relative path gets updated correctly.
        subfolder_count = len(src_file.relative_to(self.root_path).parts) - 1
        return HTML_TPL.format(
            title=src_file.stem,
            body=body,
            css_path="../" * subfolder_count + self._css_path_rel_doc,
        )

    def __call__(self, src, dst, *args, **kwargs):
        # Has same signature as shutil.copy(), for use with shutil.copytree().
        src = Path(src)
        if not self.convert_md or src.suffix.lower() != ".md":
            return shutil.copy2(src, dst, *args, **kwargs)
        dst = Path(dst)
        dst_filename = src.stem + ".htm"
        if dst.is_dir():
            dst = dst / dst_filename
        else:
            dst = dst.with_name(dst_filename)
        htm = self.md_to_html(src)
        with dst.open("w", encoding="utf8") as f:
            f.write(htm)
        self.converted_files[src] = dst
        return dst


def get_expected_help_index(fpkg_metadata: FMEPackageMetadata, format_directions=None):
    """
    Returns a mapping of the expected help contexts, and a mapping of optional help contexts,
    to the name of their corresponding HTML doc files,
    based on the formats and transformers in the given package metadata.

    Help contexts are keys used by FME Workbench to map help buttons in the UI
    to its corresponding page of documentation.
    The names of the expected HTML files follow a naming convention based on the help context.
    """
    # /foundation/ui/core/include/ui/core/helpkeywords.h
    index = {}
    # Optional help contexts
    opt = {}
    if not format_directions:
        format_directions = {}
    fpkg_ident = f"{fpkg_metadata.publisher_uid}_{fpkg_metadata.uid}"
    # Each transformer has only one topic.
    for xformer in fpkg_metadata.transformers:
        index[f"fmx_{fpkg_ident}_{xformer.name}"] = f"/{xformer.name}.htm"
    # Each format has many topics.
    for fmt in fpkg_metadata.formats:
        # dashes in the UIDs get turned to underscores when looking for format help
        fpkg_ident = fpkg_ident.replace("-", "_")
        fmt_ident = f"{fpkg_ident}_{fmt.name}".lower()
        directions = format_directions.get(fmt.name, "rw")
        # Format prefix is "rw" even when read-only or write-only
        index[f"rw_{fmt_ident}_index"] = f"/{fmt.name}.htm"
        index[f"rw_{fmt_ident}_feature_rep"] = f"/{fmt.name}_feature_rep.htm"
        for direction in directions:
            index[f"param_{fmt_ident}_{direction}"] = f"/{fmt.name}_param_{direction}.htm"
            index[f"ft_{fmt_ident}_param_{direction}"] = f"/{fmt.name}_ft_param_{direction}.htm"
        index[f"ft_{fmt_ident}_user_attr"] = f"/{fmt.name}_ft_user_attr.htm"
        # Optional help contexts
        opt[f"rw_{fmt_ident}_quickfacts"] = f"/{fmt.name}_quickfacts.htm"
    return index, opt


class HelpBuilder:
    def __init__(
        self,
        fpkg_metadata: FMEPackageMetadata,
        help_src_dir,
        help_dst_dir,
        format_visible_directions,
    ):
        self.fpkg_metadata = fpkg_metadata
        self.help_src_dir = Path(help_src_dir)
        self.help_dst_dir = Path(help_dst_dir)
        self.format_directions = format_visible_directions

    def build(self):
        """
        Copies help from src to dst. If dst folder exists, it's deleted first.

        If the src has a package_help.csv, then switch to the "manual" mode
        where doc is copied as-is and package_help.csv is validated for correctness.

        If the src doesn't have package_help.csv, then all .md files are converted to .htm
        at the dst, and generate package_help.csv at the dst.
        To map help contexts to their doc, the Markdown files must follow a naming convention.
        """
        src_index_file = self.help_src_dir / "package_help.csv"
        copier = DocCopier(self.help_src_dir)
        for item in self.fpkg_metadata.transformers:
            copier.transformer_names.add(item.name)
        if src_index_file.is_file():
            self.validate_index(self.help_src_dir)
            copier.convert_md = False

        if self.help_dst_dir.exists():
            shutil.rmtree(self.help_dst_dir)

        shutil.copytree(
            self.help_src_dir,
            self.help_dst_dir,
            ignore=shutil.ignore_patterns(*TREE_COPY_IGNORE_GLOBS),
            copy_function=copier,
        )
        if not (self.help_dst_dir / "package_help.csv").is_file():
            self.build_index(self.help_dst_dir)
        self.validate_index(self.help_dst_dir)

    def build_index(self, doc_dir):
        """
        Write package_help.csv based on package metadata and HTML help files present in doc_dir.
        To map help contexts to their doc, the Markdown files must follow a naming convention.
        Help contexts missing a doc file results in a warning,
        and its row omitted from the output.
        """
        expected, optional = get_expected_help_index(self.fpkg_metadata, self.format_directions)

        path = Path(doc_dir) / "package_help.csv"
        with path.open("w", encoding="utf8", newline="") as f:
            writer = csv.writer(f)
            for context in sorted(expected.keys()):
                expected_doc = Path(doc_dir) / expected[context][1:]  # Strip leading /
                if not expected_doc.is_file():
                    warnings.warn(f"Missing doc {expected_doc} ({context})")
                    continue
                writer.writerow([context, expected[context]])
            for context in sorted(optional.keys()):
                possible_doc = Path(doc_dir) / optional[context].lstrip("/")
                if possible_doc.is_file():
                    writer.writerow([context, optional[context]])

        return path

    def validate_index(self, doc_dir):
        """

        Read package_help.csv and validates its contents. Checks that:
        - All expected help contexts are present, based on package metadata
        - No unrecognized help contexts are present
        - Referenced files exist and are HTML or MD

        """

        links = {}
        with (Path(doc_dir) / "package_help.csv").open("r", encoding="utf8", newline="") as f:
            try:
                for row in csv.reader(f):
                    links[row[0]] = row[1]
            except IndexError as e:
                raise IndexError("Invalid package_help.csv: must have 2 columns") from e
        for ctx, doc_path in links.items():
            if not doc_path.startswith("/"):
                raise ValueError(f"Path must start with /: {doc_path}")
            expected_doc = Path(doc_dir) / doc_path.lstrip("/")
            if not expected_doc.exists():
                raise FileNotFoundError(f"{expected_doc} does not exist")
            if expected_doc.suffix[1:].lower() not in ("htm", "html", "md"):
                raise ValueError(f"{expected_doc} must be htm(l) or md")

        required, optional = get_expected_help_index(self.fpkg_metadata, self.format_directions)
        all_expected = set(required.keys()) | set(optional.keys())

        contexts_present = set(links.keys())
        unrecognized = contexts_present - all_expected
        if unrecognized:
            raise ValueError(f"Unrecognized help: {', '.join(unrecognized)}")
        missing = set(required.keys()) - contexts_present
        if missing:
            raise ValueError(f"Missing help: {', '.join(missing)}")



class TransformerPythonCompatError(ValueError):
    """Exception raised when PYTHON_COMPATIBILITY is set to unsupported version for Transformers."""

    def __init__(self, transformer_name):
        message = "'{}' Python compatibility must be >=36.".format(transformer_name)
        super().__init__(message)


class CustomTransformerPythonCompatError(ValueError):
    """Exception raised when PYTHON_COMPATIBILITY is set to unsupported version for Custom Transformers."""

    def __init__(self, transformer_name):
        message = "'{}' Python compatibility must be '2or3' or >=36.".format(transformer_name)
        super().__init__(message)



import pathlib

import pytest
from click.testing import CliRunner

from fme_packager.cli import verify, pack

CWD = pathlib.Path(__file__).parent.resolve()


def test_verify_valid():
    runner = CliRunner()
    result = runner.invoke(
        verify, [str(CWD / "fixtures" / "fpkgs" / "example.my-package-0.1.0.fpkg")]
    )
    assert result.exit_code == 0
    assert "Success" in result.output


@pytest.mark.parametrize("package_name", ["valid_package", "fmxj_package"])
def test_pack_verify(package_name):
    runner = CliRunner()
    # create a fpkg
    result = runner.invoke(pack, [str(CWD / "fixtures" / package_name)])
    assert result.exit_code == 0

    # verify the built fpkg
    dist_dir = CWD / "fixtures" / package_name / "dist"
    result = runner.invoke(verify, [str(CWD / dist_dir / "example.my-package-0.1.0.fpkg")])
    assert result.exit_code == 0
    assert "Success" in result.output


@pytest.mark.parametrize("flags", [[], ["--json"], ["--verbose"], ["--json", "--verbose"]])
def test_verify_invalid(flags):
    runner = CliRunner()
    result = runner.invoke(
        verify, [str(CWD / "fixtures" / "fpkgs" / "example.invalid-package-0.1.0.fpkg"), *flags]
    )

    if "--json" in flags:
        assert '{"status": "error", "message": "MyInvalidGreeter is missing doc"}' in result.output
    else:
        assert "MyInvalidGreeter is missing doc" in result.output

    if "--verbose" in flags:
        assert "Working on transformer" in result.output
    else:
        assert "Working on transformer" not in result.output


def test_verify_non_fpkg():
    runner = CliRunner()
    result = runner.invoke(verify, [str(CWD / "fixtures" / "valid_package" / "package.yml")])
    assert "The file must exist and have a .fpkg extension" in result.output


def test_verify_non_existent():
    runner = CliRunner()
    result = runner.invoke(verify, [str(CWD / "fixtures" / "does-not-exist.fpkg")])
    assert "The file must exist and have a .fpkg extension" in result.output



import tempfile
from json import dumps as json_dumps
from pathlib import Path

from fme_packager.operations import extract_fpkg
from fme_packager.packager import FMEPackager


class FMEVerifier:
    def __init__(self, file, verbose=False, output_json=False):
        self.file = Path(file)
        self.verbose = verbose
        self.output_json = output_json

    def verify(self):
        success, message = True, "valid"
        try:
            self._unzip_and_build()
        except Exception as e:
            success, message = False, str(e)

        if self.output_json:
            result = json_dumps({"status": "success" if success else "error", "message": message})
        else:
            result = "Success: Package Valid" if success else f"Error Validating Package: {message}"

        return result

    def _unzip_and_build(self):
        if not self.file.is_file() or not self.file.suffix.lower() == ".fpkg":
            raise ValueError("The file must exist and have a .fpkg extension")

        with tempfile.TemporaryDirectory() as temp_dir:
            extract_fpkg(self.file, temp_dir)

            # Verify the fpkg files by building the package
            steps = FMEPackager(temp_dir, self.verbose)
            steps.build()
            steps.make_fpkg()

    def _print(self, msg):
        if self.verbose:
            print(msg)



from pathlib import Path

import pytest

from fme_packager.web_service import _web_service_path, _parse_web_service

CWD = Path(__file__).parent.resolve()


@pytest.fixture
def valid_web_service():
    def web_service(name):
        return {
            "auth_type": "0",
            "connection_description": "<p>A basic connection description.</p>",
            "description": "<p>A simple web service</p>",
            "help_url": "<a href=https://example.com>my_alias</a>",
            "markdown_connection_description": "A basic connection description.",
            "markdown_description": "A simple web service",
            "service_name": f"example.my-package.{name}",
            "type": "BASIC",
            "verify_ssl_certificate_default": "true",
            "version": "1",
        }

    return web_service


def test__web_service_path():
    assert _web_service_path("test_webservice.xml") == Path("web_services") / "test_webservice.xml"


@pytest.mark.parametrize(
    "path, service_name",
    [
        (
            CWD / "fixtures" / "valid_package" / "web_services" / "My Web Service.xml",
            "My Web Service",
        ),
        (CWD / "fixtures" / "web_services" / "2024 Web Service.xml", "2024 Web Service"),
    ],
)
def test__parse_web_service(path, service_name, valid_web_service):
    assert _parse_web_service(path) == valid_web_service(service_name)


def test__parse_web_service_invalid():
    path = CWD / "fixtures" / "web_services" / "Invalid Web Service.xml"
    assert _parse_web_service(path) == {}



"""
Tools for extracting key information out of FME's transformer definition files.
"""

import json
import os.path
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Iterable, List
import csv
from io import StringIO


def parse_data_processing_type(data_processing_type: dict):
    """
    Parse data processing type from conditional logic dictionary.

    :param data_processing_type: The data processing type dict with conditional logic
    :return: List of data processing types
    """
    types = set()

    # Extract all "then" values from conditional statements
    if_conditions = data_processing_type.get("if", [])
    for condition in if_conditions:
        then_value = condition.get("then")
        if then_value:
            types.add(then_value)

    # Add default value if present
    default_value = data_processing_type.get("default")
    if default_value:
        types.add(default_value)

    return sorted(list(types))


def csv_split(line: str) -> List[str]:
    """
    Split a CSV line into a list of values, handling quoted strings.

    :param line: The CSV line to split.
    :return: List of values from the CSV line.
    """
    try:
        return next(csv.reader(StringIO(line.strip())))
    except StopIteration:
        return []


class Transformer(ABC):
    """Represents one version of a transformer."""

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def version(self) -> int:
        pass

    @property
    @abstractmethod
    def python_compatibility(self) -> str:
        pass

    @property
    @abstractmethod
    def categories(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def aliases(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def visible(self) -> bool:
        pass

    @property
    @abstractmethod
    def data_processing_types(self) -> List[str]:
        pass


@dataclass
class CustomTransformerHeader:
    name: str
    version: int
    category: List[str]
    guid: str
    insert_mode: str
    blocked_looping: str
    process_count: str
    process_group_by: str
    process_groups_ordered: str
    build_num: int
    preserves_attrs: str
    deprecated: bool
    pyver: str
    preserve_group_attr: str = ""
    replaced_by: str = ""
    data_processing_type: List[str] = field(default_factory=list)

    # noinspection PyUnreachableCode
    def __post_init__(self):
        if not isinstance(self.version, int):
            self.version = int(self.version)
        if not isinstance(self.build_num, int):
            self.build_num = int(self.build_num)
        if isinstance(self.category, str):
            self.category = csv_split(self.category)
        if isinstance(self.deprecated, str):
            self.deprecated = self.deprecated.lower() == "yes"
        if isinstance(self.data_processing_type, str):
            self.data_processing_type = self.data_processing_type.split()


def parse_custom_transformer_header(line: str) -> CustomTransformerHeader:
    """
    Parses custom transformer header.

    :param line: Custom transformer header line from FMX.
    :return: Parsed header
    """
    fields = csv_split(line.replace("# TRANSFORMER_BEGIN", ""))
    return CustomTransformerHeader(*fields)


class CustomTransformer(Transformer):
    def __init__(self, lines: List[bytes]):
        super().__init__()
        self.header: CustomTransformerHeader
        self.lines = lines
        for line in lines:
            if line.startswith(b"# TRANSFORMER_BEGIN"):
                self.header = parse_custom_transformer_header(line.decode("utf8"))
                break
        else:
            raise ValueError("TRANSFORMER_BEGIN line not found")

    @property
    def name(self):
        return self.header.name

    @property
    def version(self):
        return self.header.version

    @property
    def python_compatibility(self):
        return self.header.pyver

    @property
    def categories(self):
        return self.header.category

    @property
    def aliases(self):
        return []

    @property
    def visible(self):
        return not self.header.deprecated

    @property
    def data_processing_types(self):
        return self.header.data_processing_type

    @property
    def is_encrypted(self):
        return self.lines[0].strip() == b"FMW0001"


class FmxTransformer(Transformer):
    def __init__(self, lines):
        super().__init__()
        self.lines = lines
        self.props = {}
        for line in self.lines:
            match = re.match(r"^(.+?):\s+(.+?)$", line)
            if match:
                name = match.group(1).strip()
                if name.startswith("PARAMETER_"):
                    continue
                self.props[name] = match.group(2).strip()
        if self.name is None or self.version is None:
            raise ValueError("TRANSFORMER_NAME or VERSION not found")

    @property
    def name(self):
        return self.props.get("TRANSFORMER_NAME")

    @property
    def version(self):
        return int(self.props.get("VERSION"))

    @property
    def python_compatibility(self):
        return self.props.get("PYTHON_COMPATIBILITY")

    def _split_prop(self, property_name, sep=","):
        return (
            [p.strip() for p in self.props.get(property_name).split(sep)]
            if self.props.get(property_name)
            else []
        )

    @property
    def categories(self):
        return self._split_prop("CATEGORY", ",")

    @property
    def aliases(self):
        return self._split_prop("ALIASES", " ")

    @property
    def visible(self):
        return self.props.get("VISIBLE", "yes").lower() == "yes"

    @property
    def data_processing_types(self):
        data_processing_type = self.props.get("DATA_PROCESSING_TYPE")
        if not data_processing_type:
            return []

        data_processing_type = data_processing_type.strip()

        # Try to parse as JSON first
        try:
            parsed_data = json.loads(data_processing_type)
            return parse_data_processing_type(parsed_data)
        except json.JSONDecodeError:
            pass

        # If JSON parsing fails, treat as simple string
        return [data_processing_type]


class FmxjTransformer(Transformer):
    def __init__(self, info, version_def):
        self.info = info
        self.json_def = version_def

    @property
    def name(self):
        return self.info["name"]

    @property
    def version(self):
        return self.json_def["version"]

    @property
    def python_compatibility(self):
        # FIXME: key typo from tstportConfiguration/testdata/PortConfiguration.fmxj
        return self.json_def.get("pythonCompatability")

    @property
    def categories(self):
        return self.info.get("categories", [])

    @property
    def aliases(self):
        return self.info.get("aliases", [])

    @property
    def visible(self):
        return not self.info.get("deprecated", False)

    @property
    def data_processing_types(self):
        data_processing_type = self.json_def.get("dataProcessingType")
        if not data_processing_type:
            return []

        # Handle simple string case
        if isinstance(data_processing_type, str):
            return [data_processing_type]

        # Handle conditional logic case
        if isinstance(data_processing_type, dict):
            return parse_data_processing_type(data_processing_type)

        return []


class TransformerFile(ABC):
    """Represents a transformer file containing one or more transformer versions."""

    def __init__(self, file_path):
        self.file_path = file_path

    @abstractmethod
    def versions(self) -> Iterable[Transformer]:
        pass


def get_matching_indexes(lines, matcher):
    """
    Returns indexes of lines where line is (or can be decoded as) UTF-8 and matcher function is true.
    """
    indexes = []
    for i, line in enumerate(lines):
        if not isinstance(line, str):
            try:
                line = line.decode("utf8")
            except (UnicodeDecodeError, AttributeError):
                continue
        if matcher(line):
            indexes.append(i)
    return indexes


class FmxFile(TransformerFile):
    def __init__(self, file_path):
        super().__init__(file_path)

    def versions(self):
        with open(self.file_path, "r") as f:
            lines = f.readlines()
        transformer_begin_indexes = get_matching_indexes(
            lines, lambda l: l.startswith("TRANSFORMER_NAME")
        )
        for def_num, i in enumerate(transformer_begin_indexes):
            if i == transformer_begin_indexes[-1]:
                end_idx = len(lines) - 1
            else:
                end_idx = transformer_begin_indexes[def_num + 1] - 1
            yield FmxTransformer(lines[i:end_idx])


class CustomTransformerFmxFile(TransformerFile):
    def __init__(self, file_path):
        super().__init__(file_path)

    def versions(self):
        with open(self.file_path, "rb") as f:
            lines = f.readlines()
        transformer_begin_indexes = get_matching_indexes(
            lines, lambda l: l.startswith("# TRANSFORMER_BEGIN")
        )
        for def_num, i in enumerate(transformer_begin_indexes):
            if i == transformer_begin_indexes[-1]:
                end_idx = len(lines) - 1
            else:
                end_idx = transformer_begin_indexes[def_num + 1] - 2
            yield CustomTransformer(lines[i - 1 : end_idx])


class FmxjFile(TransformerFile):
    def __init__(self, file_path):
        super().__init__(file_path)
        with open(file_path) as f:
            self.body = json.load(f)

    def versions(self):
        for item in self.body["versions"]:
            yield FmxjTransformer(self.body["info"], item)


def load_transformer(transformer_path) -> TransformerFile:
    filename, ext = os.path.splitext(transformer_path)
    ext = ext.lower()
    if ext == ".fmx":
        # Check first line for Custom Transformer text
        with open(transformer_path, "rb") as f:
            line = f.readline()
        if line.startswith(b"#!") or line.startswith(b"FMW0001"):
            return CustomTransformerFmxFile(transformer_path)
        return FmxFile(transformer_path)
    if ext == ".fmxj":
        return FmxjFile(transformer_path)
    raise ValueError(f"Unrecognized transformer: {transformer_path}")



"""
Add FME_HOME to the Windows DLL search path or OS library load path.

See `fme-packager config-env` for details.
"""

import os
import sysconfig


# First line from fme_env.pth is FME_HOME
site_packages_dir = sysconfig.get_paths()["purelib"]
src_pth = os.path.join(site_packages_dir, "fme_env.pth")
with open(src_pth, "r") as f:
    fme_home = f.readline().strip("#\n ")

if fme_home:
    if hasattr(os, "add_dll_directory"):
        os.add_dll_directory(fme_home)
    else:
        os.environ["PATH"] = fme_home + ";" + os.environ["PATH"]



from pathlib import Path

import pytest

from fme_packager.transformer import (
    load_transformer,
    CustomTransformerFmxFile,
    CustomTransformer,
    FmxFile,
    FmxjFile,
    parse_custom_transformer_header,
)


CWD = Path(__file__).parent.resolve()


def test_custom_transformer(custom_package_dir):
    fmx = load_transformer(Path(custom_package_dir) / "transformers" / "customFooBar.fmx")
    assert isinstance(fmx, CustomTransformerFmxFile)
    defs = list(fmx.versions())
    assert len(defs) == 1
    for item in defs:
        assert isinstance(item, CustomTransformer)
        assert item.header.insert_mode == "Linked Always"
        assert item.name == "example.my-package.customFooBar"
        assert item.version == 1
        assert item.python_compatibility == "37"
        assert item.categories == ["Attributes"]
        assert item.aliases == []
        assert item.visible
        assert item.data_processing_types == []


def test_custom_transformer_encrypted_versioned(custom_package_dir):
    fmx = load_transformer(Path(custom_package_dir) / "transformers" / "customEncrypted2Ver.fmx")
    assert isinstance(fmx, CustomTransformerFmxFile)
    defs = list(fmx.versions())
    assert len(defs) == 2
    for i, item in enumerate(defs):
        assert isinstance(item, CustomTransformer)
        assert item.header.insert_mode == "Embedded Always"
        assert item.name == "example.my-package.test"
        assert item.version == len(defs) - i
        assert item.python_compatibility == "310"
        assert item.is_encrypted == (i == 0)  # Latest version is encrypted
        assert item.visible
        assert item.data_processing_types == []


@pytest.mark.parametrize(
    "line",
    [
        '# TRANSFORMER_BEGIN example.my-package.customFooBar,1,Attributes,4d874d87-714f-41b4-9dfe-9748ebd8b565,"Linked Always",No,NO_PARALLELISM,,No,19238,YES,No,37',
        '# TRANSFORMER_BEGIN customCategoriesTest,1,"Attributes,""Format Specific""",92a7f1bd-938a-49db-92c4-48fb37426eb3,"Linked Always",No,NO_PARALLELISM,,No,25741,YES,No,,No,,source',
        '# TRANSFORMER_BEGIN example.data-processing-types.bothLinked,1,,,"Linked Always",No,NO_PARALLELISM,,No,25560,YES,No,,No,,both',
    ],
    ids=["simple", "multi-category", "no-category"],
)
def test_parse_custom_transformer_header(line):
    header = parse_custom_transformer_header(line)
    assert isinstance(header.version, int)
    assert isinstance(header.build_num, int)
    assert isinstance(header.deprecated, bool)
    assert header.insert_mode == "Linked Always"

    assert header.category != [""]
    if header.name == "customCategoriesTest":
        assert header.category == ["Attributes", "Format Specific"]

    assert header.data_processing_type != [""]
    if header.name == "example.data-processing-types.bothLinked":
        assert header.data_processing_type == ["both"]


def test_fmx_transformer(valid_package_dir):
    fmx = load_transformer(Path(valid_package_dir) / "transformers" / "MyGreeter.fmx")
    assert isinstance(fmx, FmxFile)
    defs = list(fmx.versions())
    assert len(defs) == 1
    item = defs[0]
    assert item.name == "example.my-package.MyGreeter"
    assert item.version == 1
    assert item.python_compatibility == "36"
    assert item.categories == ["Integrations"]
    assert item.aliases == [
        "example.my-package.HerGreeter",
        "example.my-package.TheirGreeter",
        "example.my-package.OurGreeter",
    ]
    assert item.data_processing_types == []


def test_fmxj_transformer(fmxj_package_dir):
    fmx = load_transformer(Path(fmxj_package_dir) / "transformers" / "DemoGreeter.fmxj")
    assert isinstance(fmx, FmxjFile)
    defs = list(fmx.versions())
    assert len(defs) == 1
    item = defs[0]
    assert item.name == "example.my-package.DemoGreeter"
    assert item.version == 1
    assert not item.python_compatibility
    assert item.categories == ["Integrations"]
    assert item.aliases == ["example.my-package.NemoGreeter", "example.my-package.MemoGreeter"]
    assert item.data_processing_types == []


def test_fmx_transformer_data_processing_types(data_processing_types_package_dir):
    fmx = load_transformer(
        Path(data_processing_types_package_dir) / "transformers" / "AttributeFileReader.fmx"
    )
    assert isinstance(fmx, FmxFile)
    defs = list(fmx.versions())
    assert len(defs) == 6  # Multiple versions in the file

    for item in defs:
        assert hasattr(item, "data_processing_types")
        assert item.data_processing_types == ["source"]

    # Test destination transformer
    fmx = load_transformer(
        Path(data_processing_types_package_dir) / "transformers" / "AttributeFileWriter.fmx"
    )
    assert isinstance(fmx, FmxFile)
    defs = list(fmx.versions())
    assert len(defs) == 6  # Multiple versions in the file

    for item in defs:
        assert hasattr(item, "data_processing_types")
        assert item.data_processing_types == ["destination"]


def test_fmxj_transformer_data_processing_types(data_processing_types_package_dir):
    """Test that FMXJ transformers correctly parse conditional dataProcessingType logic."""
    fmx = load_transformer(
        Path(data_processing_types_package_dir) / "transformers" / "HTTPCaller.fmxj"
    )
    assert isinstance(fmx, FmxjFile)
    defs = list(fmx.versions())
    assert len(defs) == 8

    for item in defs:
        assert hasattr(item, "data_processing_types")
        # HTTPCaller has conditional logic with both "source" and "destination" values
        assert item.data_processing_types == ["destination", "source"]


def test_custom_transformer_data_processing_types(data_processing_types_package_dir):
    """Test that custom transformers handle missing data_processing_types gracefully."""
    fmx = load_transformer(
        Path(data_processing_types_package_dir) / "transformers" / "bothLinked.fmx"
    )
    assert isinstance(fmx, CustomTransformerFmxFile)
    defs = list(fmx.versions())
    assert len(defs) == 1

    item = defs[0]
    assert hasattr(item, "data_processing_types")
    assert item.data_processing_types == ["both"]


def test_fmx_transformer_conditional_data_processing_types(data_processing_types_package_dir):
    """Test that FMX transformers correctly parse conditional dataProcessingType logic stored as JSON."""
    fmx = load_transformer(
        Path(data_processing_types_package_dir) / "transformers" / "BoxConnector.fmx"
    )
    assert isinstance(fmx, FmxFile)
    defs = list(fmx.versions())
    assert len(defs) == 3

    for item in defs:
        assert hasattr(item, "data_processing_types")
        assert sorted(item.data_processing_types) == ["destination", "source"]



"""
fme_packager command line interface.
"""

import inspect
import json as json_lib
import os
import shutil
import subprocess
import sys
import sysconfig
from pathlib import Path

import click
from click import BadParameter
from cookiecutter.exceptions import OutputDirExistsException, FailedHookException
from cookiecutter.main import cookiecutter

import fme_packager
from fme_packager.packager import FMEPackager
from fme_packager.summarizer import summarize_fpkg
from fme_packager.verifier import FMEVerifier


@click.group()
@click.version_option(package_name="fme_packager")
def cli():
    """
    fme_packager: The FME Packages Tool.
    """
    pass


COOKIECUTTER_TEMPLATES = {
    "transformer": "https://github.com/safesoftware/fpkg-transformer-template/archive/refs/heads/main.zip",
}


@cli.command()
@click.argument("template", type=click.Choice(sorted(COOKIECUTTER_TEMPLATES.keys())))
def init(template):
    """
    Initialize a FME Package from a template.

    The template is initialized in a subdirectory of the current directory.

    TEMPLATE -- Name of the template to use.
    """
    print("Enter the values to use for the template. Press Enter to accept the [default].")
    try:
        cookiecutter(COOKIECUTTER_TEMPLATES[template])
    except (OutputDirExistsException, FailedHookException) as e:
        print(e)  # Don't show stack trace for validation errors
        sys.exit(1)


@cli.command()
@click.argument("help_path", type=click.Path(exists=True, file_okay=True))
@click.argument("fpkg_path", type=click.Path(exists=True, file_okay=False, writable=True))
def apply_help(help_path, fpkg_path):
    """
    Import a Safe TechPubs doc export into an FME Package directory.

    This operation also converts package_aliases.flali to package_help.csv at the destination.

    HELP_PATH -- Path to a ZIP or directory of a Safe TechPubs doc export. Read only.

    FPKG_PATH -- Path to the FME Package root. Its 'help' subdirectory will be recreated.
    """
    steps = FMEPackager(fpkg_path)
    steps.apply_help(help_path)


@cli.command()
@click.argument("path", type=click.Path(exists=True, file_okay=False, writable=True))
def pack(path):
    """
    Create an .fpkg file.

    Package contents are validated during this process.
    Components not referenced by the package's metadata.yml may not be included in the resulting fpkg.

    PATH -- Path to an FME Package directory.
    """
    steps = FMEPackager(path)
    steps.build()
    steps.make_fpkg()


@cli.command()
@click.argument("file", type=click.Path(exists=False))
@click.option("--verbose", "-v", is_flag=True, help="Show build steps")
@click.option("--json", is_flag=True, help="Output result as JSON")
def verify(file, verbose, json):
    """
    Verify that a .fpkg file is valid.

    Package contents are validated during this process.

    FILE -- Path to an FME .fpkg package file.
    """
    verifier = FMEVerifier(file, verbose, json)
    result = verifier.verify()
    print(result)


@cli.command()
@click.argument("file", type=click.Path(exists=True))
def summarize(file):
    """
    Extract a JSON representation of an .fpkg file.

    FILE -- Path to an FME .fpkg package file.
    """
    result = summarize_fpkg(file)
    print(json_lib.dumps(result, indent=2))


@cli.command()
@click.option(
    "--fme-home",
    prompt=True,
    default=lambda: os.environ.get("FME_HOME"),
    show_default="FME_HOME environment variable",
    type=click.Path(exists=True, file_okay=False, writable=False),
    required=True,
    help="Path to FME installation",
)
@click.option(
    "--site-packages-dir",
    default=lambda: sysconfig.get_paths()["purelib"],
    show_default="current Python environment",
    type=click.Path(exists=True, file_okay=False, writable=True),
    required=True,
    help="Path to Python environment's site-packages",
)
@click.option(
    "--verify/--no-verify",
    is_flag=True,
    default=True,
    help="Skip verification of fmeobjects access",
)
def config_env(fme_home, site_packages_dir, verify):
    """
    Configure a Python environment for access to fmeobjects
    and the Python libraries included with FME.
    """
    fme_home = Path(fme_home)
    if not (fme_home / "fme").is_file() and not (fme_home / "fme.exe").is_file():
        raise BadParameter(f"'{fme_home}' does not contain an FME executable")

    leaf_dir = "python%s%s" % (sys.version_info.major, sys.version_info.minor)
    paths_to_add = [
        "#" + fme_home.as_posix(),  # fme_env.py reads this line
        (fme_home / "python").as_posix(),  # fmeobjects API and third-party libraries
        (fme_home / "python" / leaf_dir).as_posix(),  # third-party binary libraries
    ]

    # Internal libraries bundle on Windows since FME 2023.2 (FMEENGINE-78545)
    # Mac/Linux since FME 2025.1 (FOUNDATION-22)
    internal_libs_zip = (
        fme_home / "python" / f"fme-plugins-py{sys.version_info.major}{sys.version_info.minor}.zip"
    )
    if internal_libs_zip.is_file():
        paths_to_add.append(internal_libs_zip.as_posix())

    # Location of fmeobjects before FME 2023.2 (FMEENGINE-78545)
    legacy_fmeobjects_path = fme_home / "fmeobjects" / leaf_dir
    if legacy_fmeobjects_path.is_dir():
        paths_to_add.append(legacy_fmeobjects_path.as_posix())

    site_packages_dir = Path(site_packages_dir)
    dst_pth = site_packages_dir / "fme_env.pth"
    print(f"Writing {dst_pth}")
    with dst_pth.open("w") as f:
        for path in paths_to_add:
            f.write(path + "\n")
        f.write("import fme_env\n")

    dst_py = site_packages_dir / "fme_env.py"
    print(f"Writing {dst_py}")
    src = Path(inspect.getfile(fme_packager)).parent / "fme_env.py"
    shutil.copy(src, dst_py)

    if verify:
        # Infer location of python relative to site-packages.
        # Need to run verification in the target environment,
        # which isn't necessarily the one used to run this command.
        # Unit tests for this command create a virtualenv to test against.
        executable = None
        search_dir = site_packages_dir.resolve()

        # Walk up the directory tree to find the environment root
        while search_dir.parent != search_dir:
            candidates = ["bin/python", "python", "Scripts/python.exe", "python.exe"]
            for candidate in candidates:
                if (search_dir / candidate).is_file():
                    executable = search_dir / candidate
                    break
            if executable:
                break
            search_dir = search_dir.parent

        if executable is None:
            raise FileNotFoundError(
                f"Could not find Python executable relative to {site_packages_dir}"
            )

        print("Verifying access to fmeobjects...")
        for test_import in [
            "from fmeobjects import FMEFeature",
            "import pluginbuilder",
            "import fmewebservices",
        ]:
            try:
                subprocess.run(
                    [executable, "-c", f"import sys;sys.tracebacklimit=0;{test_import}"],
                    check=True,
                    capture_output=True,
                )
            except subprocess.CalledProcessError as e:
                print(f"{test_import}: {e.stderr.decode().strip()}")
                print("Environment configuration failed.")
                print(f"Using Python {sys.version} at {executable}")
                return sys.exit(1)

    print("\nThis Python environment is now set up for access to fmeobjects.")
    print("If the FME install location changes, re-run this script to update paths.")


if __name__ == "__main__":
    cli()



"""
Tools for generating a JSON summary of an FME Package's contents.
The output is an extension of what's in metadata.yml,
with additional properties surfaced from the package's contents.
The output is for FME Hub's use, and conforms to summarizer_spec.json.
"""

import json
import os
import tempfile
from collections import namedtuple
from importlib.resources import open_text
from pathlib import Path
from typing import Iterable, List, Union
from functools import cache

from jsonschema import validate
from jsonschema.exceptions import ValidationError

from fme_packager.operations import parse_formatinfo, extract_fpkg, load_format_line
from fme_packager.transformer import load_transformer, TransformerFile
from fme_packager.web_service import _web_service_path, _parse_web_service
from fme_packager.metadata import (
    FormatMetadata,
    WebServiceMetadata,
    TransformerMetadata,
    load_fpkg_metadata,
)


TransformerFilenames = namedtuple(
    "TransformerFilenames", ["filename", "readme_filename"], defaults=[None, None]
)

FormatFilenames = namedtuple(
    "FormatFilenames", ["filename", "readme_filename", "db_filename"], defaults=[None, None, None]
)


class Summarizer:
    def __init__(self, base_dir: Union[Path, str]):
        self.base_dir: Path = Path(base_dir)

    def transformer_filenames(self, transformer_name: str) -> TransformerFilenames:
        """
        Retrieve filenames for the transformer and its readme

        :param transformer_name: The name of the transformer to retrieve filenames for.
        :return: A tuple containing the filename and readme filename.
        """
        if not transformer_name:
            return TransformerFilenames(filename=None, readme_filename=None)

        result: dict = {}
        for ext, key in [("fmx", "filename"), ("fmxj", "filename"), ("md", "readme_filename")]:
            potential_path = self.base_dir / "transformers" / f"{transformer_name}.{ext}"
            if potential_path.exists():
                result[key] = potential_path.as_posix()

        return TransformerFilenames(**result)

    def format_filenames(self, format_name: str) -> FormatFilenames:
        """
        Retrieve filenames for the format and its readme

        :param format_name: The name of the format to retrieve filenames for.
        :return: A tuple containing the filename, db filename, and readme filename.
        """
        if not format_name:
            return FormatFilenames(filename=None, readme_filename=None, db_filename=None)

        filenames = {
            "filename": (self.base_dir / "formats" / f"{format_name}.fmf").as_posix(),
            "db_filename": (self.base_dir / "formats" / f"{format_name}.db").as_posix(),
            "readme_filename": (self.base_dir / "formats" / f"{format_name}.md").as_posix(),
        }

        for key, filename in filenames.items():
            if not Path(filename).exists():
                filenames[key] = None

        return FormatFilenames(**filenames)

    def enhance_transformer_info(
        self, transformers: Iterable[TransformerMetadata]
    ) -> Iterable[dict]:
        """
        Enhance the transformer entries in the manifest with additional information.

        :return: An iterable of transformer dicts with enhanced information
        """
        result = []
        for tm in transformers:
            filenames = self.transformer_filenames(tm.name)
            loaded_transformer = load_transformer(filenames.filename)
            entry = {
                "name": tm.name,
                "latest_version": tm.version,
            }
            entry.update(_transformer_data(loaded_transformer))
            entry.update(_content_description(filenames.readme_filename))
            result.append(entry)

        return result

    def enhance_web_service_info(
        self, web_services: Iterable[WebServiceMetadata]
    ) -> Iterable[dict]:
        """
        Enhance the web service entries in the manifest with additional information.

        :return: An iterable of web service dicts with enhanced information
        """
        result = []
        for ws in web_services:
            ws_name = ws.name
            web_service_path = self.base_dir / _web_service_path(ws_name)
            web_service_content = _parse_web_service(web_service_path)
            # Normalize name without .xml suffix
            normalized_name = ws_name[:-4] if ws_name.endswith(".xml") else ws_name

            entry = {"name": normalized_name}
            for key in [
                "help_url",
                "description",
                "markdown_description",
                "connection_description",
                "markdown_connection_description",
            ]:
                entry[key] = web_service_content.get(key, "") or ""
            result.append(entry)

        return result

    def enhance_format_info(
        self, publisher_uid: str, uid: str, formats: Iterable[FormatMetadata]
    ) -> Iterable[dict]:
        """
        Enhance the format entries in the manifest with additional information.

        :return: iterable of format dicts with enhanced information
        """
        result = []
        for fm in formats:
            short_name = fm.name
            filenames = self.format_filenames(short_name)
            format_data = _format_data(load_format_line(filenames.db_filename))
            entry: dict = {
                "short_name": short_name,
                "name": f"{publisher_uid}.{uid}.{short_name}",
            }
            entry.update(_content_description(filenames.readme_filename))
            entry.update(format_data)
            result.append(entry)

        return result


def _transformer_data(loaded_file: TransformerFile) -> dict:
    """
    Get the relevant data from the loaded transformer file.

    Output dict will have the following keys:
    - 'versions': A list of dictionaries, each containing:
        - 'name': The name of the transformer.
        - 'version': The version number of the transformer.
        - 'categories': The categories of the transformer.
        - 'aliases': The aliases of the transformer.
        - 'visible': The visibility status of the transformer.
        - 'data_processing_types': The data processing types of the transformer.

    :param loaded_file: The loaded transformer file.
    :return: The updates for the transformer dictionary.
    """
    return {
        "versions": [
            {
                "name": version.name,
                "version": version.version,
                "categories": version.categories,
                "aliases": version.aliases,
                "visible": version.visible,
                "data_processing_types": version.data_processing_types,
            }
            for version in loaded_file.versions()
        ],
    }


def _content_description(readme_filename: str) -> dict:
    """
    Get the description of the transformer or format from the readme file.

    Output dict will have keys:
    - 'description': The description of the transformer or format.
    - 'description_format': The format of the description either 'md', 'text', or 'html'.

    :param readme_filename: The filename of the readme file.
    :return: The update for the transformer or format dictionary.
    """
    try:
        with open(readme_filename, "r") as file:
            return {
                "description": file.read(),
                "description_format": "md",
            }
    except OSError:
        return {
            "description": None,
            "description_format": None,
        }


def _get_all_categories(transformers: Iterable[dict], formats: Iterable[dict]) -> List[str]:
    """
    Get the union of all the categories from the transformers.

    :param transformers: An iterable of transformer dicts
    :return: An alphabetically sorted list of all categories from the transformers.
    """
    all_categories = set()
    for transformer in transformers:
        if not transformer.get("versions", None):
            continue
        highest_version = max(
            transformer["versions"], key=lambda transformer_version: transformer_version["version"]
        )
        if not highest_version["categories"]:
            continue
        all_categories.update(highest_version["categories"])

    for fmt in formats:
        if not fmt.get("categories", None):
            continue
        all_categories.update(fmt["categories"])

    all_categories_list = list(all_categories)
    all_categories_list.sort()
    return all_categories_list


def _format_data(format_line: str) -> dict:
    """
    Get the relevant data from the loaded format file.

    Output dict will have the following keys:
    - 'visible': The visibility status of the format.
    - 'fds_info': The FDS info line of the format.
    - 'categories': The categories of the format.

    :param format_line: The loaded format line.
    :return: The updates for the format dictionary.
    """
    if format_info := parse_formatinfo(format_line):
        return {
            "fds_info": format_line,
            "visible": format_info.VISIBLE.upper() != "NO",
            "categories": (
                format_info.FORMAT_CATEGORIES.split(",") if format_info.FORMAT_CATEGORIES else []
            ),
        }

    return {"visible": None, "fds_info": None, "categories": None}


@cache
def _load_output_schema() -> dict:
    """
    Load the output schema for the summarizer in jsonschema format.

    :return: The output schema.
    """
    with open_text("fme_packager", "summarizer_spec.json") as f:
        return json.load(f)


def package_deprecated(transformers: Iterable[dict], formats: Iterable[dict]) -> bool:
    """
    Determine if the package is deprecated based on its contents.
    If the package contains transformers or formats, and the highest version of all transformers are not visible,
    and if all formats are not visible, then the package is deprecated.

    :param transformers: An iterable of transformer dicts
    :param formats: An iterable of format dicts
    :return: True if the package is deprecated, False otherwise.
    """
    if not transformers and not formats:
        return False

    is_transformer_visible = any(
        max(transformer["versions"], key=lambda v: v["version"]).get("visible", True)
        for transformer in transformers
        if transformer.get("versions")
    )

    is_format_visible = any(f.get("visible", True) for f in formats)

    return not (is_transformer_visible or is_format_visible)


def summarize_fpkg(fpkg_path: Union[str, os.PathLike]) -> dict:
    """
    Summarize the FME Package.

    The output conforms to summarizer_spec.json.

    :param fpkg_path: The path to the FME Package file (.fpkg) or an already extracted package directory.
    :return: A dict summary of the FME Package, or an error dict with keys 'status' and 'message'.
    """
    input_path = Path(fpkg_path)

    # Always create a temp directory to simplify logic; only used when input is an .fpkg file
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir = Path(temp_dir)

        if input_path.is_dir():
            working_dir = input_path
        else:
            extract_fpkg(input_path, temp_dir)
            working_dir = temp_dir

        metadata = load_fpkg_metadata(working_dir)
        manifest = metadata.dict.copy()
        ctx = Summarizer(working_dir)

        manifest["package_content"] = metadata.package_content
        manifest["package_content"]["transformers"] = ctx.enhance_transformer_info(
            metadata.transformers
        )
        manifest["package_content"]["formats"] = ctx.enhance_format_info(
            metadata.publisher_uid, metadata.uid, metadata.formats
        )
        manifest["package_content"]["web_services"] = ctx.enhance_web_service_info(
            metadata.web_services
        )
        manifest["categories"] = _get_all_categories(
            manifest["package_content"]["transformers"], manifest["package_content"]["formats"]
        )
        manifest["deprecated"] = package_deprecated(
            manifest["package_content"]["transformers"], manifest["package_content"]["formats"]
        )

    try:
        validate(manifest, _load_output_schema())
    except ValidationError as e:
        return {
            "status": "error",
            "message": f"The generated output did not conform to the schema: {e.message}",
        }

    return manifest



"""
Tests for the `fme-packager config-env` command.
"""

import os
import subprocess
import sys
import venv
from pathlib import Path

import pytest
from click.testing import CliRunner

from fme_packager.cli import config_env


def invoke(*args):
    args = args or []
    runner = CliRunner()
    return runner.invoke(config_env, args)


# Note that if FME_HOME is not set and the user is prompted for it,
# the prompt loops until a valid path is provided.


def test_fme_home_invalid_path():
    result = invoke("--fme-home", "/non/existent/path")
    assert (
        "Invalid value for '--fme-home': Directory '/non/existent/path' does not exist."
        in result.output
    )


def test_fme_home_no_exe():
    result = invoke("--fme-home", ".")
    assert "does not contain an FME executable" in result.output


leaf_dir = f"python{sys.version_info.major}{sys.version_info.minor}"


@pytest.mark.parametrize(
    "also_touch",
    [
        None,
        f"fmeobjects/{leaf_dir}/fmeobjects.pyd",
        f"python/fme-plugins-py{sys.version_info.major}{sys.version_info.minor}.zip",
    ],
)
def test_outputs(tmp_path, also_touch):
    """
    Test that fme_env.pth and fme_env.py are created with expected contents.
    """
    fme_home = tmp_path / "MOCK_FME"
    fme_home.mkdir(exist_ok=True)
    (fme_home / "fme.exe").touch()

    if also_touch:
        also_touch = fme_home / also_touch
        also_touch.parent.mkdir(parents=True)
        also_touch.touch()

    site_packages = tmp_path / "mock-site-packages"
    site_packages.mkdir(exist_ok=True)

    # This is a mock of an FME installation, so don't try to verify importable fmeobjects
    result = invoke(
        "--fme-home",
        fme_home.as_posix(),
        "--site-packages-dir",
        site_packages.as_posix(),
        "--no-verify",
    )
    assert result.exit_code == 0
    assert (site_packages / "fme_env.py").is_file()
    pth = (site_packages / "fme_env.pth").read_text().split()
    assert pth[-2:] == ["import", "fme_env"]
    pth = pth[:-2]  # Remove import statement for checking paths
    assert pth[0] == f"#{fme_home.as_posix()}"
    assert pth[1] == (fme_home / "python").as_posix()
    assert pth[2] == (fme_home / "python" / leaf_dir).as_posix()

    if not also_touch:
        assert len(pth) == 3
    elif "fmeobjects" in str(also_touch.parent):
        assert also_touch.parent.as_posix() in pth  # omit fake pyd
    elif also_touch.suffix == ".zip":
        assert also_touch.as_posix() in pth


@pytest.mark.skipif(not os.getenv("FME_HOME"), reason="No FME")
@pytest.mark.slow
def test_configured_env(tmp_path):
    """
    Test that the `fme-packager config-env` command successfully runs in a clean virtualenv,
    and that the virtualenv can then import fmeobjects afterwards.
    """
    print(f"Creating venv in {tmp_path}")
    tmp_path = Path(tmp_path)
    venv.create(tmp_path)

    if sys.platform == "win32":
        bin_dir = tmp_path / "Scripts"
        py_exe = "python.exe"
    else:
        bin_dir = tmp_path / "bin"
        py_exe = "python3"
    py_exe = (bin_dir / py_exe).as_posix()
    site_packages_dir = Path(
        subprocess.run(
            [
                py_exe,
                "-c",
                "import sysconfig;print(sysconfig.get_paths()['purelib'])",
            ],
            check=True,
            capture_output=True,
        )
        .stdout.decode()
        .strip()
    )

    result = invoke(
        "--fme-home", os.getenv("FME_HOME"), "--site-packages-dir", site_packages_dir.as_posix()
    )
    print(result.output)
    assert "failed" not in result.output.lower()
    assert result.exit_code == 0



from __future__ import division, absolute_import, print_function, unicode_literals

import json
import pathlib
import shutil
from unittest.mock import patch

import pytest
from click.testing import CliRunner
from pathlib import Path

from fme_packager import summarizer
from fme_packager import cli
from fme_packager.metadata import FMEPackageMetadata
from fme_packager.operations import extract_fpkg
from fme_packager.summarizer import (
    TransformerFilenames,
    FormatFilenames,
    package_deprecated,
    Summarizer,
)

CWD = pathlib.Path(__file__).parent.resolve()


@pytest.mark.parametrize(
    "transformer_name, expected",
    [
        # Case 1: Transformer files do not exist
        (
            "non_existing_transformer",
            TransformerFilenames(filename=None, readme_filename=None),
        ),
        # Case 2: 'name' key is null
        (
            None,
            TransformerFilenames(filename=None, readme_filename=None),
        ),
        # Case 3: Transformer .fmx file exists but .fmxj file does not
        (
            "fmx_transformer",
            TransformerFilenames(
                filename=(Path("/base/dir") / "transformers" / "fmx_transformer.fmx").as_posix(),
                readme_filename=(
                    Path("/base/dir") / "transformers" / "fmx_transformer.md"
                ).as_posix(),
            ),
        ),
        # Case 4: Transformer .fmxj file exists but .fmx file does not
        (
            "fmxj_transformer",
            TransformerFilenames(
                filename=(Path("/base/dir") / "transformers" / "fmxj_transformer.fmxj").as_posix(),
                readme_filename=(
                    Path("/base/dir") / "transformers" / "fmxj_transformer.md"
                ).as_posix(),
            ),
        ),
    ],
)
def test__transformer_filenames(transformer_name, expected, mocker):
    # Mock Path.exists as an instance method (self is the path object)
    def mock_exists(self):
        path_str = str(self)
        if "non_existing_transformer" in path_str:
            return False
        if "fmx_transformer" in path_str and path_str.endswith(".fmx"):
            return True
        if "fmxj_transformer" in path_str and path_str.endswith(".fmxj"):
            return True
        if path_str.endswith(".md"):
            return True
        return False

    mocker.patch("pathlib.Path.exists", mock_exists)

    ctx = Summarizer("/base/dir")
    result = ctx.transformer_filenames(transformer_name)
    assert result == expected


@pytest.mark.parametrize(
    "format_name, exists, expected",
    [
        (
            "test_format",
            True,
            FormatFilenames(
                filename=(Path("/base/dir") / "formats" / "test_format.fmf").as_posix(),
                db_filename=(Path("/base/dir") / "formats" / "test_format.db").as_posix(),
                readme_filename=(Path("/base/dir") / "formats" / "test_format.md").as_posix(),
            ),
        ),
        (
            "non_existing_format",
            False,
            FormatFilenames(
                filename=None,
                db_filename=None,
                readme_filename=None,
            ),
        ),
    ],
)
def test__format_filenames(format_name, exists, expected):
    with patch("pathlib.Path.exists", return_value=exists):
        ctx = Summarizer("/base/dir")
        result = ctx.format_filenames(format_name)
        assert result == expected


def test__transformer_data(mock_transformer_file, mock_transformer):
    result = summarizer._transformer_data(mock_transformer_file)

    assert result["versions"] == [
        {
            "name": mock_transformer.name,
            "version": mock_transformer.version,
            "categories": mock_transformer.categories,
            "aliases": mock_transformer.aliases,
            "visible": mock_transformer.visible,
            "data_processing_types": mock_transformer.data_processing_types,
        }
    ]


def test__format_data():
    mock_format_line = (
        "SAFE.AIRTABLE.AIRTABLE|Airtable|NONE|BOTH|BOTH|NO||NON_SPATIAL|NO|YES|YES|YES|1|1|Airtable|NO"
        "|Airtable|3D,Coordinates"
    )

    result = summarizer._format_data(mock_format_line)
    assert result == {
        "visible": True,
        "fds_info": mock_format_line,
        "categories": ["3D", "Coordinates"],
    }


@pytest.fixture
def mock_transformers(mocker):
    # Mock the TransformerFile object
    mock_transformer_file1 = mocker.Mock()
    mock_transformer_file1.versions.return_value = [mocker.Mock(categories={"cat1", "cat2"})]

    mock_transformer_file2 = mocker.Mock()
    mock_transformer_file2.versions.return_value = [mocker.Mock(categories=None)]

    mock_transformer_file3 = mocker.Mock()
    mock_transformer_file3.versions.return_value = [mocker.Mock(categories={"cat3", "cat4"})]

    mock_transformer_file4 = mocker.Mock()
    mock_transformer_file4.versions.return_value = [mocker.Mock(categories={"cat4", "cat5"})]

    return [
        {
            "versions": [
                {"version": 1, "categories": {"cat6", "cat7"}},
                {"version": 2, "categories": {"cat1", "cat2"}},
            ]
        },
        {"versions": []},
        {"versions": [{"version": 1, "categories": {"cat3", "cat4"}}]},
        {"versions": [{"version": 1, "categories": {"cat4", "cat5"}}]},
    ]


@pytest.fixture
def mock_formats():
    return [
        {"categories": {"cat-f1", "cat3"}, "deprecated": False},
        {"categories": {"cat-f2", "cat4"}, "deprecated": False},
    ]


def test__get_all_categories(mock_transformers, mock_formats):
    result = summarizer._get_all_categories(mock_transformers, mock_formats)
    assert result == ["cat-f1", "cat-f2", "cat1", "cat2", "cat3", "cat4", "cat5"]


def test__add_content_description(mocker):
    transformer = {"readme_filename": "MyGreeter.md"}
    mocker.patch("builtins.open", mocker.mock_open(read_data="Test Description"))
    result = summarizer._content_description(transformer["readme_filename"])
    assert result["description"] == "Test Description"
    assert result["description_format"] == "md"


def test_package_transformers_deprecated(mock_transformers, mock_formats):
    # Case: All formats are not visible
    mock_formats[0]["visible"] = False
    mock_formats[1]["visible"] = False
    assert not package_deprecated(mock_transformers, mock_formats)

    # Case: One transformer version is not visible
    mock_transformers[0]["versions"][1]["visible"] = False
    assert not package_deprecated(mock_transformers, mock_formats)

    # Case: Multiple transformer versions are not visible
    mock_transformers[0]["versions"][1]["visible"] = False
    mock_transformers[2]["versions"][0]["visible"] = False
    assert not package_deprecated(mock_transformers, mock_formats)

    # Case: All transformer versions are not visible
    mock_transformers[3]["versions"][0]["visible"] = False
    assert package_deprecated(mock_transformers, mock_formats)

    mock_transformers[0]["versions"][0]["visible"] = True
    assert package_deprecated(mock_transformers, mock_formats)


def test_package_formats_deprecated(mock_transformers, mock_formats):
    # Case: All transformer versions are not visible
    mock_transformers[0]["versions"][1]["visible"] = False
    mock_transformers[2]["versions"][0]["visible"] = False
    mock_transformers[3]["versions"][0]["visible"] = False
    assert not package_deprecated(mock_transformers, mock_formats)

    # Case: One format is not visible
    mock_formats[0]["visible"] = False
    assert not package_deprecated(mock_transformers, mock_formats)

    # Case: All formats are not visible
    mock_formats[1]["visible"] = False
    assert package_deprecated(mock_transformers, mock_formats)


def test_package_all_visible(mock_transformers, mock_formats):
    for transformer in mock_transformers:
        for version in transformer["versions"]:
            version["visible"] = True
    for f in mock_formats:
        f["visible"] = True
    assert not package_deprecated(mock_transformers, mock_formats)


VALID_PACKAGE_PATH = CWD / "fixtures" / "fpkgs" / "example.my-package-0.1.0.fpkg"


@pytest.mark.parametrize(
    "fpkg_path, expected_output_path",
    [
        (
            VALID_PACKAGE_PATH,
            CWD / "fixtures" / "json_output" / "summarize_example.my-package-0.1.0.fpkg.json",
        ),
        (
            CWD / "fixtures" / "fpkgs" / "example.my-format-0.1.0.fpkg",
            CWD / "fixtures" / "json_output" / "summarize_example.my-format-0.1.0.fpkg.json",
        ),
        (
            CWD / "fixtures" / "fpkgs" / "example.my-package-deprecated-0.1.0.fpkg",
            CWD
            / "fixtures"
            / "json_output"
            / "summarize_example.my-package-deprecated-0.1.0.fpkg.json",
        ),
        (
            CWD / "fixtures" / "fpkgs" / "example.data-processing-types-0.1.0.fpkg",
            CWD
            / "fixtures"
            / "json_output"
            / "summarize_example.data-processing-types-0.1.0.fpkg.json",
        ),
    ],
    ids=["simple", "format", "deprecated", "data_processing_types"],
)
def test_summarize_fpkg(fpkg_path, expected_output_path):
    expected_output = json.loads(expected_output_path.read_text())
    result = summarizer.summarize_fpkg(fpkg_path)
    assert result == expected_output


MOCK_EMPTY_METADATA = {
    "fpkg_version": 1,
    "uid": "my-package",
    "publisher_uid": "example",
    "name": "My FME Package",
    "description": "A short description of my FME Package.",
    "version": "0.1.0",
    "minimum_fme_build": 19238,
    "author": {"name": "G Raymond", "email": "me@example.com"},
}
with open(CWD / "fixtures" / "json_output" / "summarize_example.empty.json") as f:
    EXPECTED_MOCK_EMPTY_SUMMARIZE_OUTPUT = json.load(f)


def test_summarize_empty_fpkg(monkeypatch, tmp_path):
    """
    Valiate summary output for an FPKG with no package_content.
    Start with a valid .fpkg and mock out its metadata.
    """
    mock_metadata = FMEPackageMetadata(MOCK_EMPTY_METADATA)
    monkeypatch.setattr(summarizer, "load_fpkg_metadata", lambda x: mock_metadata)
    result = summarizer.summarize_fpkg(VALID_PACKAGE_PATH)
    assert result == EXPECTED_MOCK_EMPTY_SUMMARIZE_OUTPUT


def test_summarize_unsupported_from_build(monkeypatch, tmp_path):
    """
    Valiate summary output for an FPKG with unsupported_from_build set.
    Use the empty metadta mock, and add unsupported_from_build to it.
    """
    mock_metadata_dict = MOCK_EMPTY_METADATA.copy()  # shallow is ok
    mock_metadata_dict["unsupported_from_build"] = 26000
    mock_metadata = FMEPackageMetadata(mock_metadata_dict)
    assert mock_metadata.unsupported_from_build == 26000
    monkeypatch.setattr(summarizer, "load_fpkg_metadata", lambda x: mock_metadata)
    expected_output = EXPECTED_MOCK_EMPTY_SUMMARIZE_OUTPUT.copy()  # shallow is ok
    expected_output["unsupported_from_build"] = 26000
    result = summarizer.summarize_fpkg(VALID_PACKAGE_PATH)
    assert result == expected_output


def test_summarize():
    runner = CliRunner()
    fpkg_path = str(CWD / "fixtures" / "fpkgs" / "example.my-package-0.1.0.fpkg")
    result = runner.invoke(cli.summarize, [fpkg_path])
    assert result.exit_code == 0
    json.loads(result.output)


@pytest.fixture
def fpkg_dir_with_bad_data_processing_type(tmp_path):
    tmp_path = Path(tmp_path)
    fpkg_path = CWD / "fixtures" / "fpkgs" / "example.my-package-0.1.0.fpkg"
    extract_fpkg(fpkg_path, tmp_path)

    # Modify the FMX file to have an invalid DATA_PROCESSING_TYPE
    fmx_path = tmp_path / "transformers" / "MyGreeter.fmx"
    fmx_txt = fmx_path.read_text()
    fmx_txt = fmx_txt.replace("ALIASES:", "ALIASES:\nDATA_PROCESSING_TYPE: bad value\n")
    fmx_path.write_text(fmx_txt)
    return tmp_path


def test_summarize_bad_data_processing_type(fpkg_dir_with_bad_data_processing_type):
    """
    Test that summarize_fpkg can run against an extracted FPKG,
    and rejects a bad data_processing_type in the FMX.
    """
    result = summarizer.summarize_fpkg(fpkg_dir_with_bad_data_processing_type)
    assert result["status"] == "error"
    assert result["message"].startswith(
        "The generated output did not conform to the schema: 'bad value' is not one of"
    )


@pytest.mark.parametrize(
    "subcommand", [cli.summarize, cli.verify, cli.pack], ids=lambda func: str(func)
)
def test_cli_bad_data_processing_type(fpkg_dir_with_bad_data_processing_type, subcommand):
    """
    Test that CLI commands fail with a bad data_processing_type in the FMX.
    """
    if subcommand is cli.verify:
        # Re-zip the extracted package back to .fpkg.
        # Can't use `fme-packager pack` because it's expected to reject this bad package.
        dest = str(fpkg_dir_with_bad_data_processing_type / "example.my-package-0.1.0.fpkg")
        shutil.make_archive(
            dest,
            "zip",
            root_dir=fpkg_dir_with_bad_data_processing_type,
        )
        Path(dest + ".zip").rename(dest)
        args = [dest]
    else:
        args = [fpkg_dir_with_bad_data_processing_type.as_posix()]

    result = CliRunner().invoke(subcommand, args)
    if subcommand is cli.pack:
        assert "The generated output did not conform to the schema" in str(result)
        assert result.exit_code != 0
    else:
        assert "The generated output did not conform to the schema" in result.output
        assert result.exit_code == 0
